# -*- coding: utf-8 -*-
"""This module defines functions for parsing header data from PDB files."""

from collections import defaultdict, OrderedDict
import numpy as np
import os.path

from prody import LOGGER
from prody.atomic import flags, AAMAP
from prody.atomic.atomic import invAAMAP
from prody.utilities import openFile, alignBioPairwise, GAP_EXT_PENALTY

from .localpdb import fetchPDB
from .header import (Chemical, Polymer, DBRef, _PDB_DBREF,
                     cleanString)

from .starfile import parseSTARSection

__all__ = ['parseCIFHeader', 'getCIFHeaderDict']

_COMPND_KEY_MAPPINGS = {'_entity.id': 'MOL_ID',
                        '_entity.pdbx_description': 'MOLECULE',
                        '_entity.pdbx_fragment': 'FRAGMENT',
                        '_entity_name_com.name': 'SYNONYM',
                        '_entity.pdbx_ec': 'EC',
                        '_entity.pdbx_mutation': 'MUTATION',
                        '_entity.details': 'OTHER_DETAILS'}

_JRNL___KEY_MAPPINGS = {'_citation_author.name': 'AUTH',
                        '_citation.title': 'TITL',
                        #'_citation_editor.name': 'EDIT', # not normally found
                        '_citation.journal_abbrev': 'REF',
                        '_citation.journal_volume': 'REF',
                        '_citation.page_first': 'REF',
                        #'_citation.page_last': 'REF', # not in pdb
                        '_citation.year': 'REF',
                        #'_entity.pdbx_ec': 'PUBL', # publisher not supported
                        #'_citation.journal_id_ASTM': 'REFN',
                        '_citation.journal_id_ISSN': 'REFN',
                        #'_citation.journal_id_ISBN': 'REFN',
                        '_citation.pdbx_database_id_PubMed': 'PMID',
                        '_citation.pdbx_database_id_DOI': 'DOI'}

def _natomsFromFormula(formula, hydrogens=False):
    if ("(" in formula and ")" in formula):
        formula = formula.split("(")[1].split(")")[0]

    parts = formula.split()
    if not hydrogens:
        parts = [part for part in parts if part.find("H") == -1]

    return sum([_natomsFromFormulaPart(part) for part in parts])


def _natomsFromFormulaPart(part):
    digits = [s for s in part if s.isdigit()]
    if len(digits) == 0:
        return 1
    return int("".join(digits))

def parseCIFHeader(pdb, *keys, **kwargs):
    """Returns header data dictionary for *pdb*.  This function is equivalent to
    ``parsePDB(pdb, header=True, model=0, meta=False)``, likewise *pdb* may be
    an identifier or a filename.

    List of header records that are parsed.

    ============ ================= ============================================
    Record type  Dictionary key(s)  Description
    ============ ================= ============================================
    HEADER       | classification  | molecule classification
                 | deposition_date | deposition date
                 | identifier      | PDB identifier
    TITLE        title             title for the experiment or analysis
    SPLIT        split             list of PDB entries that make up the whole
                                   structure when combined with this one
    COMPND       polymers          see :class:`Polymer`
    EXPDTA       experiment        information about the experiment
    NUMMDL       n_models          number of models
    MDLTYP       model_type        additional structural annotation
    AUTHOR       authors           list of contributors
    JRNL         reference         reference information dictionary:
                                     * *authors*: list of authors
                                     * *title*: title of the article
                                     * *editors*: list of editors
                                     * *issn*:
                                     * *reference*: journal, vol, issue, etc.
                                     * *publisher*: publisher information
                                     * *pmid*: pubmed identifier
                                     * *doi*: digital object identifier
    DBREF[1|2]   polymers          see :class:`Polymer` and :class:`DBRef`
    SEQADV       polymers          see :class:`Polymer`
    SEQRES       polymers          see :class:`Polymer`
    MODRES       polymers          see :class:`Polymer`
    HELIX        polymers          see :class:`Polymer`
    SHEET        polymers          see :class:`Polymer`
    HET          chemicals         see :class:`Chemical`
    HETNAM       chemicals         see :class:`Chemical`
    HETSYN       chemicals         see :class:`Chemical`
    FORMUL       chemicals         see :class:`Chemical`
    REMARK   2   resolution        resolution of structures, when applicable
    REMARK   4   version           PDB file version
    REMARK 350   biomoltrans       biomolecular transformation lines
                                   (unprocessed)
    REMARK 900	 related_entries   related entries in the PDB or EMDB
    ============ ================= ============================================

    Header records that are not parsed are: OBSLTE, CAVEAT, SOURCE, KEYWDS,
    REVDAT, SPRSDE, SSBOND, LINK, CISPEP, CRYST1, ORIGX1, ORIGX2, ORIGX3,
    MTRIX1, MTRIX2, MTRIX3, and REMARK X not mentioned above."""

    if not os.path.isfile(pdb):
        if len(pdb) == 4 and pdb.isalnum():
            filename = fetchPDB(pdb, format='cif', compressed=False)
            if filename is None:
                raise IOError('PDB file for {0} could not be downloaded.'
                              .format(pdb))
            pdb = filename
        else:
            raise IOError('{0} is not a valid filename or a valid PDB '
                          'identifier.'.format(pdb))
    pdb = openFile(pdb, 'rt')
    header = getCIFHeaderDict(pdb, *keys, **kwargs)
    pdb.close()
    return header


def getCIFHeaderDict(stream, *keys, **kwargs):
    """Returns header data in a dictionary.  *stream* may be a list of PDB lines
    or a stream.
    
    Polymers have sequences that usually use one-letter residue name abbreviations by default. 
    To obtain long (usually three letter) abbrevations, set *longSeq* to **True**.
    """

    try:
        lines = stream.readlines()
        stream.close()
    except:
        lines = stream

    pdbid = _PDB_HEADER_MAP['identifier'](lines)
    if keys:
        keys = list(keys)
        for k, key in enumerate(keys):
            if key in _PDB_HEADER_MAP:
                if key == 'polymers':
                    value = _PDB_HEADER_MAP[key](lines, **kwargs)
                else:
                    value = _PDB_HEADER_MAP[key](lines)
                keys[k] = value
            else:
                try:
                    if key == 'polymers':
                        value = _PDB_HEADER_MAP[key](lines, **kwargs)
                    else:
                        value = _PDB_HEADER_MAP[key](lines)
                    keys[k] = value
                except:
                    raise KeyError('{0} is not a valid header data identifier'
                                .format(repr(key)))
            if key in ('chemicals', 'polymers'):
                for component in value:
                    component.pdbentry = pdbid
        if len(keys) == 1:
            return keys[0]
        else:
            return tuple(keys)
    else:
        header = {}
        for key, func in _PDB_HEADER_MAP.items():  # PY3K: OK
            value = func(lines)
            if value is not None:
                header[key] = value
        for chem in header.get('chemicals', []):
            chem.pdbentry = pdbid
            header[chem.resname] = chem
        for poly in header.get('polymers', []):
            poly.pdbentry = pdbid
            header[poly.chid] = poly
        return header


def _getBiomoltrans(lines):

    
    biomolecule = defaultdict(list)

    # 2 blocks are needed for this:
    # _pdbx_struct_assembly_gen: what to apply to which chains
    # _pdbx_struct_oper_list: everything else
    data1 = parseSTARSection(lines, '_pdbx_struct_assembly_gen', report=False)
    data2 = parseSTARSection(lines, '_pdbx_struct_oper_list', report=False)

    # extracting the data
    for n, item1 in enumerate(data1):
        currentBiomolecule = item1["_pdbx_struct_assembly_gen.assembly_id"]
        applyToChains = []

        chains = item1["_pdbx_struct_assembly_gen.asym_id_list"].replace(';','').strip().split(',')
        applyToChains.extend(chains)

        biomt = biomolecule[currentBiomolecule]

        oper_expression = item1["_pdbx_struct_assembly_gen.oper_expression"]
        if oper_expression[0].isnumeric():
            operators = oper_expression.split(',')
        elif (oper_expression.startswith('(')
              and oper_expression.find('-') != -1
              and oper_expression.endswith(')')):
            firstOperator = int(oper_expression.split('(')[1].split('-')[0])-1
            lastOperator = int(oper_expression.split('-')[1].split(')')[0])
            operators = range(firstOperator, lastOperator)
        else:
            operators = []
        for oper in operators:
            biomt.append(applyToChains)

            item2 = data2[int(oper)-1]

            for i in range(1,4):
                biomt.append(" ".join([
                    item2["_pdbx_struct_oper_list.matrix[{0}][1]".format(i)],
                    item2["_pdbx_struct_oper_list.matrix[{0}][2]".format(i)],
                    item2["_pdbx_struct_oper_list.matrix[{0}][3]".format(i)],
                    item2["_pdbx_struct_oper_list.vector[{0}]".format(i)],
                ]))

    return dict(biomolecule)


def _getResolution(lines):

    for line in lines:
        if line.startswith("_reflns.d_resolution_high"):
            try:
                return float(line.split()[1])
            except:
                pass
    return None


def _getRelatedEntries(lines):

    dbrefs = []

    try:
        key = "_pdbx_database_related"
        data = parseSTARSection(lines, key, report=False)
        for item in data:
            dbref = DBRef()
            dbref.accession = item[key + ".db_id"]
            dbref.dbabbr = item[key + ".db_name"]
            dbref.database = _PDB_DBREF.get(dbref.dbabbr, 'Unknown')
            
            dbrefs.append(dbref)
    except:
        LOGGER.warn("No related entries found")
        return None

    return dbrefs


def _getSpaceGroup(lines):

    for line in lines:
        if line.startswith("_symmetry.space_group_name_H-M"):
            try:
                return line.split()[1]
            except:
                pass
    return None


def _getHelix(lines):

    alphas = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    helix = {} 
    
    i = 0
    fields = OrderedDict()
    fieldCounter = -1
    foundHelixBlock = False
    doneHelixBlock = False
    start = 0
    stop = 0
    while not doneHelixBlock and i < len(lines):
        line = lines[i]
        if line[:13] == '_struct_conf.':
            fieldCounter += 1
            fields[line.split('.')[1].strip()] = fieldCounter

        if line.startswith('HELX_P'):
            if not foundHelixBlock:
                foundHelixBlock = True
                start = i
        else:
            if foundHelixBlock:
                doneHelixBlock = True
                stop = i
        i += 1

    if i < len(lines):
        for line in lines[start:stop]:
            data = line.split()

            try:
                chid = data[fields["beg_auth_asym_id"]]
                segn = data[fields["beg_label_asym_id"]]
                value = (int(data[fields["pdbx_PDB_helix_class"]]),
                        int(data[fields["id"]][6:]),
                        data[fields["pdbx_PDB_helix_id"]].strip())
            except:
                continue

            initICode = data[fields["pdbx_beg_PDB_ins_code"]]
            if initICode == '?':
                initICode = ' '

            initResnum = int(data[fields["beg_auth_seq_id"]])
            if initICode != ' ':
                for icode in alphas[alphas.index(initICode):]:
                    helix[(chid, initResnum, icode, segn)] = value
                initResnum += 1

            endICode = data[fields["pdbx_end_PDB_ins_code"]]
            if endICode == '?':
                endICode = ' '

            endResnum = int(data[fields["end_auth_seq_id"]])
            if endICode != ' ':
                for icode in alphas[:alphas.index(endICode)+1]:
                    helix[(chid, endResnum, icode, segn)] = value
                endResnum -= 1

            for resnum in range(initResnum, endResnum+1):
                helix[(chid, resnum, '', segn)] = value

    return helix


def _getHelixRange(lines):

    alphas = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    helix = []

    i = 0
    fields = OrderedDict()
    fieldCounter = -1
    foundHelixBlock = False
    doneHelixBlock = False
    start = 0
    stop = 0
    while not doneHelixBlock and i < len(lines):
        line = lines[i]
        if line[:13] == '_struct_conf.':
            fieldCounter += 1
            fields[line.split('.')[1].strip()] = fieldCounter

        if line.startswith('HELX_P'):
            if not foundHelixBlock:
                foundHelixBlock = True
                start = i
        else:
            if foundHelixBlock:
                doneHelixBlock = True
                stop = i
        i += 1

    if i < len(lines):
        for line in lines[start:stop]:
            data = line.split()

            try:
                chid = data[fields["beg_auth_asym_id"]]
                Hclass=int(data[fields["pdbx_PDB_helix_class"]])
                Hnr=int(data[fields["id"]][6:])
            except:
                continue

            initResnum = int(data[fields["beg_auth_seq_id"]])
            endICode = data[fields["pdbx_end_PDB_ins_code"]]
            if endICode == '?':
                endICode = ' '
            endResnum = int(data[fields["end_auth_seq_id"]])
            if endICode != ' ':
                endResnum -= 1
            helix.append(['H', chid, Hclass, Hnr, initResnum, endResnum]) 
     
    return helix


def _getSheet(lines):

    alphas = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    sheet = {}

    # mmCIF files have this data divided between 4 blocks

    # block 1 has how many strands are in each sheet - col 14:16 in PDB
    i = 0
    fields1 = OrderedDict()
    fieldCounter1 = -1
    foundSheetBlock1 = False
    foundSheetBlockData1 = False
    doneSheetBlock1 = False
    start1 = 0
    stop1 = 0
    while not doneSheetBlock1 and i < len(lines):
        line = lines[i]
        if line[:14] == '_struct_sheet.':
            fieldCounter1 += 1
            fields1[line.split('.')[1].strip()] = fieldCounter1
            if not foundSheetBlock1:
                foundSheetBlock1 = True

        if foundSheetBlock1:
            if not line.startswith('#'):
                if not foundSheetBlockData1:
                    start1 = i
                    foundSheetBlockData1 = True
            else:
                if foundSheetBlockData1:
                    doneSheetBlock1 = True
                    stop1 = i

        i += 1

    n_strands_dict = {}

    if i < len(lines):
        for line in lines[start1:stop1]:
            data = line.split()

            try:
                sheet_id = data[fields1["id"]]
                num_strands = int(data[fields1["number_strands"]])
            except:
                continue

            n_strands_dict[sheet_id] = num_strands

    # block 2 has the packing order and sense direction - col 38:40
    i = 0
    fields2 = OrderedDict()
    fieldCounter2 = -1
    foundSheetBlock2 = False
    foundSheetBlockData2 = False
    doneSheetBlock2 = False
    start2 = 0
    stop2 = 0
    while not doneSheetBlock2 and i < len(lines):
        line = lines[i]
        if line[:20] == '_struct_sheet_order.':
            fieldCounter2 += 1
            fields2[line.split('.')[1].strip()] = fieldCounter2
            if not foundSheetBlock2:
                foundSheetBlock2 = True

        if foundSheetBlock2:
            if not line.startswith('#'):
                if not foundSheetBlockData2:
                    start2 = i
                    foundSheetBlockData2 = True
            else:
                if foundSheetBlockData2:
                    doneSheetBlock2 = True
                    stop2 = i

        i += 1

    sense_dict = {}
    
    if i < len(lines):
        for sheet_id in n_strands_dict:
            sense_dict[sheet_id, 1] = 0

        for line in lines[start2:stop2]:
            data = line.split()

            try:
                sheet_id = data[fields2["sheet_id"]]
                strand2 = int(data[fields2["range_id_2"]])
                sense = data[fields2["sense"]]
            except:
                continue

            if sense == "parallel":
                sense_dict[sheet_id, strand2] = 1
            else:
                sense_dict[sheet_id, strand2] = -1

    # block 3 has the ranges - col 7:37 - NB: sheet and strand ID switched
    i = 0
    fields3 = OrderedDict()
    fieldCounter3 = -1
    foundSheetBlock3 = False
    foundSheetBlockData3 = False
    doneSheetBlock3 = False
    start3 = 0
    stop3 = 0
    while not doneSheetBlock3 and i < len(lines):
        line = lines[i]
        if line[:20] == '_struct_sheet_range.':
            fieldCounter3 += 1
            fields3[line.split('.')[1].strip()] = fieldCounter3
            if not foundSheetBlock3:
                foundSheetBlock3 = True

        if foundSheetBlock3:
            if not line.startswith('#'):
                if not foundSheetBlockData3:
                    start3 = i
                    foundSheetBlockData3 = True
            else:
                if foundSheetBlockData3:
                    doneSheetBlock3 = True
                    stop3 = i

        i += 1

    if i < len(lines):
        for line in lines[start3:stop3]:
            data = line.split()

            try:
                chid = data[fields3["beg_auth_asym_id"]]
                segn = data[fields3["beg_label_asym_id"]]

                strand_id = int(data[fields3["id"]])
                sheet_id = data[fields3["sheet_id"]]
                sense = sense_dict[sheet_id, strand_id]
                value = (sense, strand_id, sheet_id)
            except:
                continue

            initICode = data[fields3["pdbx_beg_PDB_ins_code"]]
            if initICode == '?':
                initICode = ' '

            initResnum = int(data[fields3["beg_auth_seq_id"]])
            if initICode != ' ':
                for icode in alphas[alphas.index(initICode):]:
                    sheet[(chid, initResnum, icode, segn)] = value
                initResnum += 1

            endICode = data[fields3["pdbx_end_PDB_ins_code"]]
            if endICode == '?':
                endICode = ' '

            endResnum = int(data[fields3["end_auth_seq_id"]])
            if endICode != ' ':
                for icode in alphas[:alphas.index(endICode)+1]:
                    sheet[(chid, endResnum, icode, segn)] = value
                endResnum -= 1

            for resnum in range(initResnum, endResnum+1):
                sheet[(chid, resnum, '', segn)] = value

    # block 4 has the hydrogen bonds - col 41:70 - NB: very different order
    # we don't actually use this info though

    return sheet


def _getSheetRange(lines):

    alphas = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    sheet = []

    # mmCIF files have this data divided between 4 blocks

    # block 1 has how many strands are in each sheet - col 14:16 in PDB
    i = 0
    fields1 = OrderedDict()
    fieldCounter1 = -1
    foundSheetBlock1 = False
    foundSheetBlockData1 = False
    doneSheetBlock1 = False
    start1 = 0
    stop1 = 0
    while not doneSheetBlock1 and i < len(lines):
        line = lines[i]
        if line[:14] == '_struct_sheet.':
            fieldCounter1 += 1
            fields1[line.split('.')[1].strip()] = fieldCounter1
            if not foundSheetBlock1:
                foundSheetBlock1 = True

        if foundSheetBlock1:
            if not line.startswith('#'):
                if not foundSheetBlockData1:
                    start1 = i
                    foundSheetBlockData1 = True
            else:
                if foundSheetBlockData1:
                    doneSheetBlock1 = True
                    stop1 = i

        i += 1

    n_strands_dict = {}
    if i < len(lines):
        for line in lines[start1:stop1]:
            data = line.split()

            try:
                sheet_id = data[fields1["id"]]
                num_strands = int(data[fields1["number_strands"]])
            except:
                continue

            n_strands_dict[sheet_id] = num_strands

    # block 2 has the packing order and sense direction - col 38:40
    i = 0
    fields2 = OrderedDict()
    fieldCounter2 = -1
    foundSheetBlock2 = False
    foundSheetBlockData2 = False
    doneSheetBlock2 = False
    start2 = 0
    stop2 = 0
    while not doneSheetBlock2 and i < len(lines):
        line = lines[i]
        if line[:20] == '_struct_sheet_order.':
            fieldCounter2 += 1
            fields2[line.split('.')[1].strip()] = fieldCounter2
            if not foundSheetBlock2:
                foundSheetBlock2 = True

        if foundSheetBlock2:
            if not line.startswith('#'):
                if not foundSheetBlockData2:
                    start2 = i
                    foundSheetBlockData2 = True
            else:
                if foundSheetBlockData2:
                    doneSheetBlock2 = True
                    stop2 = i

        i += 1

    sense_dict = {}
    if i < len(lines):
        for sheet_id in n_strands_dict:
            sense_dict[sheet_id, 1] = 0

        for line in lines[start2:stop2]:
            data = line.split()

            try:
                sheet_id = data[fields2["sheet_id"]]
                strand2 = int(data[fields2["range_id_2"]])
                sense = data[fields2["sense"]]
            except:
                continue

            if sense == "parallel":
                sense_dict[sheet_id, strand2] = 1
            else:
                sense_dict[sheet_id, strand2] = -1

    # block 3 has the ranges - col 7:37 - NB: sheet and strand ID switched
    i = 0
    fields3 = OrderedDict()
    fieldCounter3 = -1
    foundSheetBlock3 = False
    foundSheetBlockData3 = False
    doneSheetBlock3 = False
    start3 = 0
    stop3 = 0
    while not doneSheetBlock3 and i < len(lines):
        line = lines[i]
        if line[:20] == '_struct_sheet_range.':
            fieldCounter3 += 1
            fields3[line.split('.')[1].strip()] = fieldCounter3
            if not foundSheetBlock3:
                foundSheetBlock3 = True

        if foundSheetBlock3:
            if not line.startswith('#'):
                if not foundSheetBlockData3:
                    start3 = i
                    foundSheetBlockData3 = True
            else:
                if foundSheetBlockData3:
                    doneSheetBlock3 = True
                    stop3 = i

        i += 1

    if i < len(lines):
        for line in lines[start3:stop3]:
            data = line.split()

            try:
                chid = data[fields3["beg_auth_asym_id"]]

                Snr = int(data[fields3["id"]])
                sheet_id = data[fields3["sheet_id"]]
                dir = sense_dict[sheet_id, Snr]
            except:
                continue

            initICode = data[fields3["pdbx_beg_PDB_ins_code"]]
            if initICode == '?':
                initICode = ' '

            initResnum = int(data[fields3["beg_auth_seq_id"]])
            if initICode != ' ':
                initResnum += 1

            endICode = data[fields3["pdbx_end_PDB_ins_code"]]
            if endICode == '?':
                endICode = ' '

            endResnum = int(data[fields3["end_auth_seq_id"]])
            if endICode != ' ':
                endResnum -= 1

            sheet.append(['E', chid, dir, Snr, initResnum, endResnum])

    return sheet


def _getReference(lines):
    """Returns a reference of the PDB entry."""

    ref = {}
    title = ''
    authors = []
    editors = []
    reference = ''
    publisher = ''

    # JRNL double block. Blocks 6 and 7 as copied from COMPND
    # Block 1 has most info. Block 2 has author info
    items1 = parseSTARSection(lines, "_citation", report=False)
    items2 = parseSTARSection(lines, "_citation_author", report=False)

    for row in items1:
        for k, value in row.items():
            try:
                what = _JRNL___KEY_MAPPINGS[k]
            except:
                continue
            if what == 'TITL':
                title += value.upper()
            elif what == 'REF':
                if k == "_citation.journal_volume":
                    reference += " V. "
                reference += ' ' + value.upper()
                reference = reference.strip()
            elif what == 'REFN':
                ref['issn'] = 'ISSN ' + value
            elif what == 'PMID':
                ref['pmid'] = value
            elif what == 'DOI':
                ref['doi'] = value.upper()

    for row in items2:
        for k, value in row.items(): 
            try:
                what = _JRNL___KEY_MAPPINGS[k]
            except:
                continue
            if what == 'AUTH':
                try:
                    surname, initials = value.split(',')
                    author = initials+surname
                except ValueError:
                    author = value
                authors.append(author.strip().upper())

        ref['authors'] = authors
        ref['title'] = cleanString(title)
        ref['editors'] = editors
        ref['reference'] = cleanString(reference)
        ref['publisher'] = cleanString(publisher)

    return ref


def _getPolymers(lines, **kwargs):
    """Returns list of polymers (macromolecules).
    
    Sequence is usually one-letter abbreviations, but can be long 
    abbreviations (usually three letters) if *longSeq* is **True**"""

    pdbid = _PDB_HEADER_MAP['identifier'](lines)
    polymers = dict()

    entities = defaultdict(list)

    # SEQRES block
    items1 = parseSTARSection(lines, '_entity_poly', report=False)

    for item in items1:
        chains = item.get('_entity_poly.pdbx_strand_id', '').replace(';','').replace(' ', '')
        entity = item.get('_entity_poly.entity_id', '')

        for ch in chains.split(","):
            entities[entity].append(ch)
            poly = polymers.get(ch, Polymer(ch))
            polymers[ch] = poly

            longSeq = kwargs.get('longSeq', False)
            if longSeq:
                poly.sequence += ''.join(item[
                    '_entity_poly.pdbx_seq_one_letter_code'].replace(';', '').split())
            else:
                poly.sequence += ''.join(item[
                    '_entity_poly.pdbx_seq_one_letter_code_can'].replace(';', '').split())

    if longSeq:
        for poly in polymers.values():
            seq = poly.sequence
            resnames = []
            for item in seq.split('('):
                if item.find(')') != -1:
                    resnames.append(item[:item.find(')')])
                    letters = list(item[item.find(')')+1:])
                else:
                    letters = list(item)
                resnames.extend([invAAMAP[letter] for letter in letters])

            poly.sequence = ' '.join(resnames)

    # DBREF block 1
    items2 = parseSTARSection(lines, '_struct_ref', report=False)

    for item in items2:
        entity = item["_struct_ref.id"]
        chains = entities[entity]
        for ch in chains:
            dbabbr = item["_struct_ref.db_name"]
            dbref = DBRef()
            dbref.dbabbr = dbabbr
            dbref.database = _PDB_DBREF.get(dbabbr, 'Unknown')
            dbref.accession = item["_struct_ref.pdbx_db_accession"]
            dbref.idcode = item["_struct_ref.db_code"]

            poly = polymers[ch]
            poly.dbrefs.append(dbref)

    # DBREF block 2
    items3 = parseSTARSection(lines, "_struct_ref_seq", report=False)

    for i, item in enumerate(items3):
        i += 1

        ch = item["_struct_ref_seq.pdbx_strand_id"]
        poly = polymers[ch] 
        
        for dbref in poly.dbrefs:
            if item["_struct_ref_seq.pdbx_db_accession"] == dbref.accession:

                try:
                    first = int(item["_struct_ref_seq.pdbx_auth_seq_align_beg"])
                    initICode = item["_struct_ref_seq.pdbx_db_align_beg_ins_code"]
                    if initICode == '?':
                        initICode = ' '
                except:
                    try:
                        first = int(item["_struct_ref_seq.pdbx_auth_seq_align_beg"])
                        initICode = item["_struct_ref_seq.pdbx_seq_align_beg_ins_code"]
                        if initICode == '?':
                            initICode = ' '
                    except:
                        LOGGER.warn('DBREF for chain {2}: failed to parse '
                                    'initial sequence number of the PDB sequence '
                                    '({0}:{1})'.format(pdbid, i, ch))
                try:
                    last = int(item["_struct_ref_seq.pdbx_auth_seq_align_end"])
                    endICode = item["_struct_ref_seq.pdbx_db_align_end_ins_code"]
                    if endICode == '?':
                        endICode = ' '
                except:
                    try:
                        last = int(item["_struct_ref_seq.pdbx_auth_seq_align_end"])
                        endICode = item["_struct_ref_seq.pdbx_seq_align_end_ins_code"]
                        if endICode == '?':
                            endICode = ' '
                    except:
                        LOGGER.warn('DBREF for chain {2}: failed to parse '
                                    'ending sequence number of the PDB sequence '
                                    '({0}:{1})'.format(pdbid, i, ch))
                try:
                    first2 = int(item["_struct_ref_seq.db_align_beg"])
                    dbref.first = (first, initICode, first2)
                except:
                    LOGGER.warn('DBREF for chain {2}: failed to parse '
                                'initial sequence number of the database sequence '
                                '({0}:{1})'.format(pdbid, i, ch))
                try:
                    last2 = int(item["_struct_ref_seq.db_align_end"])
                    dbref.last = (last, endICode, last2)
                except:
                    LOGGER.warn('DBREF for chain {2}: failed to parse '
                                'ending sequence number of the database sequence '
                                '({0}:{1})'.format(pdbid, i, ch))

    for poly in polymers.values():  # PY3K: OK
        resnum = []
        for dbref in poly.dbrefs:
            dbabbr = dbref.dbabbr
            if dbabbr == 'PDB':
                if not (pdbid == dbref.accession == dbref.idcode):
                    LOGGER.warn('DBREF for chain {2} refers to PDB '
                                'entry {3} ({0}:{1})'
                                .format(pdbid, i, ch, dbref.accession))
            else:
                if pdbid == dbref.accession or pdbid == dbref.idcode:
                    LOGGER.warn('DBREF for chain {2} is {3}, '
                                'expected PDB ({0}:{1})'
                                .format(pdbid, i, ch, dbabbr))
                    dbref.database = 'PDB'
            
            try:
                resnum.append((dbref.first[0], dbref.last[0]))
            except:
                pass # we've already warned about this

        resnum.sort()
        last = -10000
        for first, temp in resnum:
            if first <= last:
                LOGGER.warn('DBREF records overlap for chain {0} ({1})'
                            .format(poly.chid, pdbid))
            last = temp

    # MODRES block
    data4 = parseSTARSection(lines, "_pdbx_struct_mod_residue", report=False)

    for data in data4:
        ch = data["_pdbx_struct_mod_residue.label_asym_id"]

        poly = polymers.get(ch, Polymer(ch))
        polymers[ch] = poly
        if poly.modified is None:
            poly.modified = []

        iCode = data["_pdbx_struct_mod_residue.PDB_ins_code"]
        if iCode == '?':
            iCode == '' # PDB one is stripped
        poly.modified.append((data["_pdbx_struct_mod_residue.auth_comp_id"],
                                data["_pdbx_struct_mod_residue.auth_asym_id"],
                                data["_pdbx_struct_mod_residue.auth_seq_id"] + iCode,
                                data["_pdbx_struct_mod_residue.parent_comp_id"],
                                data["_pdbx_struct_mod_residue.details"]))

    # SEQADV block
    data5 = parseSTARSection(lines, "_struct_ref_seq_dif", report=False)

    for i, data in enumerate(data5):
        ch = data["_struct_ref_seq_dif.pdbx_pdb_strand_id"]

        poly = polymers.get(ch, Polymer(ch))
        polymers[ch] = poly
        dbabbr = data["_struct_ref_seq_dif.pdbx_seq_db_name"]
        resname = data["_struct_ref_seq_dif.mon_id"]
        if resname == '?':
            resname = '' # strip for pdb

        try:
            resnum = int(data["_struct_ref_seq_dif.pdbx_auth_seq_num"])
        except:
            #LOGGER.warn('SEQADV for chain {2}: failed to parse PDB sequence '
            #            'number ({0}:{1})'.format(pdbid, i, ch))
            continue

        icode = data["_struct_ref_seq_dif.pdbx_pdb_ins_code"]
        if icode == '?':
            icode = '' # strip for pdb            
        
        try:
            dbnum = int(data["_struct_ref_seq_dif.pdbx_seq_db_seq_num"])
        except:
            #LOGGER.warn('SEQADV for chain {2}: failed to parse database '
            #            'sequence number ({0}:{1})'.format(pdbid, i, ch))
            continue            

        comment = data["_struct_ref_seq_dif.details"].upper()
        if comment == '?':
            comment = '' # strip for pdb 

        match = False
        for dbref in poly.dbrefs:
            if dbref.first is None or dbref.last is None:
                continue
            if not dbref.first[0] <= resnum <= dbref.last[0]:
                continue
            match = True
            if dbref.dbabbr != dbabbr:
                LOGGER.warn('SEQADV for chain {2}: reference database '
                            'mismatch, expected {3} parsed {4} '
                            '({0}:{1})'.format(pdbid, i+1, ch,
                            repr(dbref.dbabbr), repr(dbabbr)))
                continue
            dbacc = data["_struct_ref_seq_dif.pdbx_seq_db_accession_code"]
            if dbref.accession[:9] != dbacc[:9]:
                LOGGER.warn('SEQADV for chain {2}: accession code '
                            'mismatch, expected {3} parsed {4} '
                            '({0}:{1})'.format(pdbid, i+1, ch,
                            repr(dbref.accession), repr(dbacc)))
                continue
            dbref.diff.append((resname, resnum, icode, dbnum, dbnum, comment))
        if not match:
            LOGGER.warn('SEQADV for chain {2}: database sequence reference '
                        'not found ({0}:{1})'.format(pdbid, i+1, ch))
            continue

    # COMPND double block. 
    # Block 6 has most info. Block 7 has synonyms
    data6 = parseSTARSection(lines, "_entity", report=False)
    data7 = parseSTARSection(lines, "_entity_name_com", report=False)

    dict_ = {}
    for molecule in data6:
        dict_.clear()
        for k, value in molecule.items():
            if k == '_entity.id':
                dict_['CHAIN'] = ', '.join(entities[value])

            try:
                key = _COMPND_KEY_MAPPINGS[k]
            except:
                continue
            val = value.strip()
            if val == '?':
                val = ''
            dict_[key.strip()] = val

        chains = dict_.pop('CHAIN', '').strip()

        if not chains:
            continue
        for ch in chains.split(','):
            ch = ch.strip()
            poly = polymers.get(ch, Polymer(ch))
            polymers[ch] = poly
            poly.name = dict_.get('MOLECULE', '').upper()

            poly.fragment = dict_.get('FRAGMENT', '').upper()

            poly.comments = dict_.get('OTHER_DETAILS', '').upper()

            val = dict_.get('EC', '')
            poly.ec = [s.strip() for s in val.split(',')] if val else []

            poly.mutation = dict_.get('MUTATION', '') != ''
            poly.engineered = dict_.get('ENGINEERED', poly.mutation)

    for molecule in data7:
        dict_.clear()
        for k, value in molecule.items():
            if k.find('entity_id') != -1:
                dict_['CHAIN'] = ', '.join(entities[value])

            try:
                key = _COMPND_KEY_MAPPINGS[k]
            except:
                continue
            dict_[key.strip()] = value.strip()

        chains = dict_.pop('CHAIN', '').strip()

        if not chains:
            continue
        for ch in chains.split(','):
            ch = ch.strip()
            poly = polymers.get(ch, Polymer(ch))
            polymers[ch] = poly

            val = dict_.get('SYNONYM', '')
            poly.synonyms = [s.strip().upper() for s in val.split(',')
                                ] if val else []

    return list(polymers.values())


def _getChemicals(lines):
    """Returns list of chemical components (heterogens)."""

    chemicals = defaultdict(list)
    chem_names = defaultdict(str)
    chem_synonyms = defaultdict(str)
    chem_formulas = defaultdict(str)
    chem_n_atoms = defaultdict(int)

    # Data is split across blocks again

    # 1st block we need is has info about location in structure

    # this instance only includes single sugars not branched structures
    items = parseSTARSection(lines, "_pdbx_nonpoly_scheme", report=False)

    for data in items:
        resname = data["_pdbx_nonpoly_scheme.mon_id"]
        if resname in flags.AMINOACIDS or resname == "HOH":
            continue

        chem = Chemical(resname)
        chem.chain = data["_pdbx_nonpoly_scheme.pdb_strand_id"]
        chem.resnum = int(data["_pdbx_nonpoly_scheme.pdb_seq_num"])

        icode = data["_pdbx_nonpoly_scheme.pdb_ins_code"]
        if icode == '.':
            icode = ''
        chem.icode = icode
        chem.description = '' # often empty in .pdb and not clearly here
        chemicals[chem.resname].append(chem)

    # next we get the equivalent one for branched sugars part
    items = parseSTARSection(lines, "_pdbx_branch_scheme", report=False)

    for data in items:
        resname = data["_pdbx_branch_scheme.mon_id"]
        if resname in flags.AMINOACIDS or resname == "HOH":
            continue

        chem = Chemical(resname)
        chem.chain = data["_pdbx_branch_scheme.pdb_asym_id"]
        chem.resnum = int(data["_pdbx_branch_scheme.pdb_seq_num"])

        chem.icode = '' # this part doesn't have this field
        chem.description = '' # often empty in .pdb and not clearly here
        chemicals[chem.resname].append(chem)

    # 2nd block to get has general info e.g. name and formula
    items = parseSTARSection(lines, "_chem_comp", report=False)

    for data in items:
        resname = data["_chem_comp.id"]
        if resname in flags.AMINOACIDS or resname == "HOH":
            continue

        chem_names[resname] += data["_chem_comp.name"].upper()

        if "_chem_comp.pdbx_synonyms" in data.keys():
            synonym = data["_chem_comp.pdbx_synonyms"]
        else:
            synonym = '?'

        if synonym == '?':
            synonym = ' '
        synonym = synonym.rstrip()
        if synonym.startswith(';') and synonym.endswith(';'):
            synonym = synonym[1:-1]
        chem_synonyms[resname] += synonym
        
        if "_chem_comp.formula" in data.keys():
            chem_formulas[resname] += data["_chem_comp.formula"]

    for key, name in chem_names.items():  # PY3K: OK
        name = cleanString(name)
        for chem in chemicals[key]:
            chem.name = name

    for key, formula in chem_formulas.items():  # PY3K: OK
        formula = cleanString(formula)
        repeats = len(chemicals[key])
        for chem in chemicals[key]:
            chem.formula = '{0}({1})'.format(repeats, formula)
            chem.natoms = _natomsFromFormula(formula)

    for key, synonyms in chem_synonyms.items():  # PY3K: OK
        synonyms = cleanString(synonyms)
        synonyms = synonyms.split(';')
        for chem in chemicals[key]:
            if synonyms != ['']:
                chem.synonyms = [syn.strip() for syn in synonyms]

    alist = []
    for chem in chemicals.values():  # PY3K: OK
        for chem in chem:
            alist.append(chem)
    return alist


def _getVersion(lines):

    for line in lines:
        if line.startswith("_audit_conform.dict_version"):
            return line.split()[1]
    return None


def _getNumModels(lines):

    for line in lines:
        if line.startswith("_pdbx_nmr_ensemble.conformers_submitted_total_number"):
            data = line.split()[1]
            try:
                return int(data)
            except:
                pass
    return None


def _getTitle(lines):

    title = ''

    try:
        data = parseSTARSection(lines, "_struct", report=False)
        for item in data:
            title += item['_struct.title'].upper()
    except:
        pass

    if len(title) == 0:
        title = None
        
    return title


def _getAuthors(lines):

    authors = []

    try:
        data = parseSTARSection(lines, "_audit_author", report=False)
        for item in data:
            author = ''.join(item['_audit_author.name'].split(', ')[::-1])
            authors.append(author.upper())
    except:
        pass

    if len(authors) == 0:
        authors = None
        
    return authors


def _getSplit(lines):

    split = []

    key = "_pdbx_database_related"

    try:
        data, _ = parseSTARSection(lines, key, report=False)
        for item in data:
            if item[key + '.content_type'] == 'split':
                split.append(item[key + '.db_id'])
    except:
        pass

    if len(split) == 0:
        split = None
        
    return split


def _getModelType(lines):

    model_type = ''

    model_type += [line.split()[1]
                  if line.find("_struct.pdbx_model_type_details") != -1 else ''
                  for line in lines][0]

    if model_type == '?':
        model_type = None

    return model_type


def _getOther(lines, key=None):

    if key is None:
        return None

    data = []

    try:
        data = parseSTARSection(lines, key, report=False)
    except:
        pass

    if len(data) == 0:
        data = None

    return data


def _getUnobservedSeq(lines, **kwargs):
    """Get sequence of unobserved residues.
    
    This sequence is usually using one-letter residue name abbreviations by default. 
    To obtain long (usually three letter) abbrevations, set *longSeq* to **True**."""

    key_unobs = '_pdbx_unobs_or_zero_occ_residues'

    unobs = []
    polymers = []
    try:
        unobs = parseSTARSection(lines, key_unobs, report=False)
        polymers = _getPolymers(lines, **kwargs)
    except:
        pass

    if len(unobs) == 0:
        return None

    unobs_seqs = OrderedDict()
    for item in unobs:
        chid = item['_pdbx_unobs_or_zero_occ_residues.auth_asym_id']
        if not chid in unobs_seqs.keys():
            unobs_seqs[chid] = ''
        unobs_seqs[chid] += AAMAP[item['_pdbx_unobs_or_zero_occ_residues.auth_comp_id']]

    if len(unobs_seqs) == 0:
        return None

    if len(polymers) == 0:
        return None

    full_seqs = OrderedDict()
    for item in polymers:
        chid = item.chid
        full_seqs[chid] = item.sequence

    if len(full_seqs) == 0:
        return None

    alns = OrderedDict()
    for _, (key, seq) in enumerate(full_seqs.items()):
        if key in unobs_seqs.keys():
            # initialise alignment with all gaps for unobs
            row1 = '-'*len(seq)
            row1_list = list(row1)
            aln = [row1, seq]

            # fix it
            for j, item in enumerate(unobs):
                chid = item['_pdbx_unobs_or_zero_occ_residues.auth_asym_id']
                if chid == key:
                    if len(item['_pdbx_unobs_or_zero_occ_residues.auth_comp_id']) == 1:
                        one_letter = item['_pdbx_unobs_or_zero_occ_residues.auth_comp_id']
                    else:
                        one_letter = AAMAP[item['_pdbx_unobs_or_zero_occ_residues.auth_comp_id']].upper()
                    good_pos = int(item['_pdbx_unobs_or_zero_occ_residues.label_seq_id']) - 1
                    row1_list = list(aln[0])
                    row1_list[good_pos] = one_letter
                    aln[0] = ''.join(row1_list)

            alns[key] = aln

    return alns


# Make sure that lambda functions defined below won't raise exceptions
_PDB_HEADER_MAP = {
    'helix': _getHelix,
    'helix_range': _getHelixRange,
    'sheet': _getSheet,
    'sheet_range': _getSheetRange,
    'chemicals': _getChemicals,
    'polymers': _getPolymers,
    'reference': _getReference,
    'resolution': _getResolution,
    'biomoltrans': _getBiomoltrans,
    'version': _getVersion,
    'deposition_date': lambda lines: [line.split()[1]
                                      if line.find("initial_deposition_date") != -1 else None
                                      for line in lines][0],
    'classification': lambda lines: [line.split()[1]
                                     if line.find("_struct_keywords.pdbx_keywords") != -1 else None
                                     for line in lines][0],
    'identifier': lambda lines: [line.split('_')[1]
                                 if line.find("data") == 0 else ''
                                 for line in lines][0].strip(),
    'title': _getTitle,
    'experiment': lambda lines: [line.split()[1]
                                 if line.find("_exptl.method") != -1 else None
                                 for line in lines][0],
    'authors': _getAuthors,
    'split': _getSplit,
    'model_type': _getModelType,
    'n_models': _getNumModels,
    'space_group': _getSpaceGroup,
    'related_entries': _getRelatedEntries,
    'others': _getOther,
    'unobserved': _getUnobservedSeq
}



# -*- coding: utf-8 -*-
"""This module defines functions for interfacing with the EBI's
Gene Ontology Annotation (GOA) database for analysing gene/protein functions 
through the Gene Ontology (GO).

This module is based on the tutorial notebook at 
https://nbviewer.jupyter.org/urls/dessimozlab.github.io/go-handbook/GO%20Tutorial%20in%20Python%20-%20Solutions.ipynb"""

from io import BytesIO
from os.path import isdir, isfile, join, split, splitext, normpath
import os
import gzip
from ftplib import FTP

from collections import defaultdict
import re
import numpy as np
from prody.utilities import makePath, gunzip, relpath, copyFile, openURL
from prody.utilities import openFile, isListLike, sympath
from prody import LOGGER, PY3K
from prody.ensemble import Ensemble

if PY3K:
    import urllib.parse as urllib
    import urllib.request as urllib2
else:
    import urllib
    import urllib2


__all__ = ['GOADictList', 'parseOBO', 'parseGAF',
           'queryGOA', 'showGoLineage',
           'calcGoOverlap', 'calcDeepFunctionOverlaps',
           'calcEnsembleFunctionOverlaps', 'findDeepestFunctions',
           'findDeepestCommonAncestor', 'calcMinBranchLength',
           'findCommonParentGoIds']


class GOADictList(object):
    """A class for handling the list of GOA Dictionaries returned 
    by queryGOA
    """

    def __init__(self, parsingList, title='unnamed', **kwargs):
        go = kwargs.pop('go', None)
        if go is None:
            go = parseOBO(**kwargs)

        self._title = title
        self._list = parsingList
        go_ids = np.unique([entry['GO_ID'] for entry in parsingList])
        self._go_terms = [go[id] for id in go_ids]
        self.numTerms = len(self._go_terms)

        self._molecular = [
            term for term in self._go_terms if term.namespace == 'molecular_function']
        self._cellular = [
            term for term in self._go_terms if term.namespace == 'cellular_component']
        self._process = [
            term for term in self._go_terms if term.namespace == 'biological_process']

    def __getitem__(self, ind):
        return self._list[ind]

    def getTitle(self):
        return self._title

    def setTitle(self, value):
        self._title = value

    def getList(self):
        return self._list

    def __repr__(self):
        if self.numTerms == 1:
            return '<GOADictList: {0} (1 GOA dict)>'.format(self._title)
        return '<GOADictList: {0} ({1} GOA dicts)>'.format(self._title, self.numTerms)

    def __iter__(self):
        """Yield go_term instances."""
        for item in self._list:
            yield item

    def pop(self, index):
        """Pop dataBlock with the given index from the list of dataBlocks in GOADictList"""
        self._go_terms.pop(index)
        self.numTerms -= 1

    def getGoTerms(self):
        return self._go_terms

    def getMolecularFunctions(self):
        return self._molecular

    def getCellularComponents(self):
        return self._cellular

    def getBiologicalProcesses(self):
        return self._process


def parseOBO(**kwargs):
    """Parse a GO OBO file containing the GO itself.
    See `OBO`_ for more information on the file format.

    .. _OBO: http://owlcollab.github.io/oboformat/doc/obo-syntax.html
    """
    try:
        from goatools import obo_parser
    except:
        raise ImportError('GOATools needs to be installed to use parseOBO')

    go_obo_url = kwargs.get('go_obo_url', None)
    if go_obo_url is None:
        go_obo_url = 'http://purl.obolibrary.org/obo/go/go-basic.obo'

    data_folder = kwargs.get('data_folder', None)
    if data_folder is None:
        data_folder = os.getcwd() + '/Data'

    # Check if we have the ./data directory already
    if(not os.path.isfile(data_folder)):
        # Emulate mkdir -p (no error if folder exists)
        try:
            os.mkdir(data_folder)
        except OSError as e:
            if(e.errno != 17):
                raise e
    else:
        raise Exception('Data path (' + data_folder + ') exists as a file. '
                        'Please rename, remove or change the desired location of the data path.')

    # Check if the file exists already
    if(not os.path.isfile(data_folder+'/go-basic.obo')):
        try:
            handle = openURL(go_obo_url)
        except Exception as err:
            raise Exception('{0} download failed ({1}).'.format(
                go_obo_url, str(err)))
        else:
            data = handle.read()
            if len(data):
                go_obo = data_folder+'/go-basic.obo'

                with open(go_obo, 'w+b') as obofile:
                    obofile.write(data)

                LOGGER.debug('{0} downloaded ({1})'
                             .format(go_obo_url, sympath(go_obo)))
            else:
                raise Exception('{0} download failed, reason unknown.'
                            .format(go_obo_url))

    else:
        go_obo = data_folder+'/go-basic.obo'

    return obo_parser.GODag(go_obo)


def parseGAF(database='PDB', **kwargs):
    """Parse a GO Association File (GAF) corresponding to
    a particular database collection into a dictionary 
    for ease of querying.

    See `GAF`_ for more information on the file format

    .. _GAF: http://geneontology.org/docs/go-annotation-file-gaf-format-2.1/

    :arg database: name of the database of interest
        default is PDB. Others include UNIPROT and 
        common names of many organisms.
    :type database: str

    :arg filename: filename for the gaf of interest
        default is goa_ and the database name in lower case
        and .gaf.gz
    :type filename: str
    """
    import Bio.UniProt.GOA as GOA

    if not isinstance(database, str):
        raise TypeError('database should be a string')

    database = database.upper()
    filename = kwargs.get('filename', None)
    if filename is None:
        if database == 'UNIPROT':
            filename = 'goa_' + database.lower() + '_all.gaf.gz'
        else:
            filename = 'goa_' + database.lower() + '.gaf'

    data_folder = kwargs.get('data_folder', os.getcwd())

    # If the file doesn't already exist, download it
    gaf = os.path.join(data_folder, filename)
    if not(os.path.exists(gaf) and os.path.getsize(gaf) > 0):
        LOGGER.info('Downloading file {0} to {1}'.format(filename, gaf))
        data_stream = BytesIO()
        ftp_host = 'ftp.ebi.ac.uk'
        ftp = FTP(ftp_host)
        ftp.login()

        try:
            ftp.cwd('pub/databases/GO/goa')
            ftp.cwd(database)
            ftp.retrbinary('RETR {}.gz'.format(filename), data_stream.write)
        except:
            raise ValueError('Cannot find the requested GO association file')

        # Logout from FTP server
        ftp.quit()

        zip_data = data_stream.getvalue()
        data_stream.close()

        rawdata = gunzip(zip_data)
        if PY3K:
            rawdata = rawdata.decode()

        with open(filename, 'w') as gaf_fp:
            gaf_fp.write(rawdata)

        LOGGER.info('Download completed for file {0}'.format(filename))

    with open(filename, 'rt') as gaf_fp:
        funcs = defaultdict(list)  # Initialise the dictionary of functions

        # Iterate on each function using Bio.UniProt.GOA library.
        LOGGER.info('Iterating through entries in {0}'.format(gaf))
        for entry in GOA.gafiterator(gaf_fp):
            id = entry.pop('DB_Object_ID')
            funcs[id].append(entry)

    return funcs


def queryGOA(*ids, **kwargs):
    """Query a GOA database by identifier.

    :arg ids: an identifier or a list-like of identifiers 
    :type ids: str, tuple, list, :class:`~numpy.ndarray`

    :arg database: name of the database of interest
        default is PDB. Others include UNIPROT and 
        common names of many organisms.
    :type database: str
    """
    database = kwargs.pop('database', 'PDB')

    gaf_dict = kwargs.pop('gaf_dict', None)
    if gaf_dict is None:
        gaf_dict = parseGAF(database=database, **kwargs)
        LOGGER.info('GAF parsing completed.')

    n_ids = len(ids)
    if n_ids == 1:
        if isListLike(ids[0]):
            ids = ids[0]
            n_ids = len(ids)

    if n_ids == 1:
        ids = list(ids)

    results = []
    unmapped = []
    LOGGER.progress('Querying GOA for {0} ids...'
                    .format(n_ids), n_ids, '_prody_queryGOA')
    for i, id in enumerate(ids):
        LOGGER.update(i, 'Querying GOA for id {0} of {1}...'
                      .format(i+1, n_ids), label='_prody_queryGOA')
        if not isinstance(id, str):
            raise TypeError('each ID should be a string')

        id = id.upper()

        if database == 'PDB':
            if not len(id) in [4, 5, 6]:
                raise ValueError('PDB IDs should be strings of length 4 to 6')

            if len(id) == 5 and str.isalpha(id[-1]):
                id = id[:4] + '_' + id[-1]

        if id in list(gaf_dict.keys()):
            results.append(gaf_dict[id])
        else:
            results.append([])
            unmapped.append(id)

    rets = []
    LOGGER.progress('Mapping GO terms back to GOA results for {0} ids...'
                    .format(n_ids), n_ids, '_prody_mapGO')
    for i, result in enumerate(results):
        LOGGER.update(i, 'Mapping GO terms back to GOA results id {0} of {1}...'
                      .format(i+1, n_ids), label='_prody_mapGO')
        rets.append(GOADictList(result, title=ids[i], **kwargs))

    if n_ids == 1:
        rets = rets[0]

    return rets


def calcGoOverlap(*go_terms, **kwargs):
    """Calculate overlap between GO terms based on their distance
    in the graph. GO terms in different namespaces (molecular function,
    cellular component, and biological process) have undefined distances.

    :arg go_terms: a list of GO terms or GO IDs
    :type go_terms: list, tuple, `~numpy.ndarray`

    :arg pairwise: whether to calculate to a matrix of pairwise overlaps
        default is False
    :type pairwise: bool

    :arg distance: whether to return distances rather than calculating overlaps
        default is False
    :type distance: bool

    :arg go: GO graph. Default behaviour is to parse it with :func:`.parseOBO`.
    :type go: `~goatools.obo_parser.GODag`
    """
    pairwise = kwargs.pop('pairwise', False)
    distance = kwargs.get('distance', False)
    operator = kwargs.get('operator', None)

    go = kwargs.get('go', None)
    if go is None:
        go = parseOBO(**kwargs)

    if not isListLike(go_terms):
        raise TypeError('please provide a list-like of go terms')

    if pairwise:
        distances = np.zeros((len(go_terms), len(go_terms)))
        for i, go_terms_i in enumerate(go_terms):
            for j, go_terms_j in enumerate(go_terms):
                distances[i, j] = calcGoOverlap(
                    go_terms_i, go_terms_j, pairwise=False, **kwargs)

    else:
        go_terms1 = go_terms[0]

        flattened_term_list = []
        for entry in go_terms[1:]:
            if isListLike(entry):
                flattened_term_list.extend(entry)
            else:
                flattened_term_list.append(entry)

        if not isListLike(go_terms1):
            go_terms1 = [go_terms1]

        if not isListLike(flattened_term_list):
            flattened_term_list = [flattened_term_list]

        try:
            flattened_term_list = [go[term] for term in flattened_term_list]
            go_terms1 = [go[term] for term in go_terms1]
        except:
            try:
                flattened_term_list = [term.id for term in flattened_term_list]
                go_terms1 = [term.id for term in go_terms1]
            except:
                raise TypeError('go_terms should contain go terms or IDs')

        for term in flattened_term_list:
            if not isinstance(term, str):
                term = term.id

        for term in go_terms1:
            if not isinstance(term, str):
                term = term.id

        distances = np.zeros((len(go_terms1), len(flattened_term_list)))
        for i, go_id1 in enumerate(go_terms1):
            for j, go_id2 in enumerate(flattened_term_list):
                distances[i, j] = calcMinBranchLength(go_id1, go_id2, go)

        if operator is not None and isListLike(distances):
            distances = operator(distances)

    if operator is None:
        if distances.shape[-1] == 1:
            distances = distances.flatten()

        if distances.shape == (1,):
            distances = distances[0]

    if distance:
        return distances
    else:
        return 1. / distances


def calcMinBranchLength(go_id1, go_id2, go):
    '''Find the minimum branch length between two terms in the GO DAG.

    :arg go_id1: the first GO ID
    :type go_id1: str

    :arg go_id2: the second GO ID
    :type go_id2:str

    :arg go: object containing a gene ontology (GO) directed acyclic graph (DAG)
    :type go: `~goatools.obo_parser.GODag`
    '''
    # First get the deepest common ancestor
    dca = findDeepestCommonAncestor([go_id1, go_id2], go)
    if dca is None:
        LOGGER.warn('There are no common ancestors between {0} and {1} so no meaningful distance can be calculated.'.format(
            go_id1, go_id2))
        return None

    # Then get the distance from the DCA to each term
    dca_depth = go[dca].depth
    d1 = go[go_id1].depth - dca_depth
    d2 = go[go_id2].depth - dca_depth

    # Return the total distance - i.e., to the deepest common ancestor and back.
    return d1 + d2


def findDeepestCommonAncestor(terms, go):
    '''Find the nearest common ancestor. 
    Only returns single most specific - assumes unique exists.

    :arg terms: a list of GO terms
    :type terms: tuple, list, :class:`~numpy.ndarray`

    :arg go: object containing a gene ontology (GO) directed acyclic graph (DAG)
    :type go: `~goatools.obo_parser.GODag`
    '''
    # Take the element at maximum depth.
    common_parent = findCommonParentGoIds(terms, go)
    if common_parent == set():
        return None
    return max(common_parent, key=lambda t: go[t].depth)


def findCommonParentGoIds(terms, go):
    '''This function finds the common ancestors in the GO 
    tree of the list of terms in the input.

    :arg terms: a list of GO terms
    :type terms: tuple, list, :class:`~numpy.ndarray`

    :arg go: object containing a gene ontology (GO) directed acyclic graph (DAG)
    :type go: `~goatools.obo_parser.GODag`
    '''
    # Find candidates from first
    rec = go[terms[0]]
    candidates = rec.get_all_parents()
    candidates.update({terms[0]})

    # Find intersection with second to nth term
    for term in terms[1:]:
        rec = go[term]
        parents = rec.get_all_parents()
        parents.update({term})

        # Find the intersection with the candidates, and update.
        candidates.intersection_update(parents)

    return candidates


def showGoLineage(go_term, **kwargs):
    """Use pygraphviz and IPython notebook to show the lineage of a GO term

    :arg go: object containing a gene ontology (GO) directed acyclic graph (DAG) 
        default is to parse with :func:`.parseOBO`
    :type go: `~goatools.obo_parser.GODag`

    arg out_format: format for output. 
        Currently only output to file. This file will be displayed in Jupyter Notebook.
    type out_format: str

    arg filename: filename for output
        default behaviour is to use the GO term ID and append '_lineage.png'
    type filename: str
    """
    out_format = kwargs.pop('format', 'png')

    if out_format == 'png':
        filename = kwargs.pop('filename', '_'.join(
            go_term.id.split(':')) + '_lineage.png')

        go = kwargs.pop('go', None)
        if go is None:
            go = parseOBO(**kwargs)

        go.draw_lineage([go_term], lineage_img=filename)
        from IPython.display import Image
        Image(filename)


def calcDeepFunctionOverlaps(*goa_data, **kwargs):
    """Calculate function overlaps between the deep 
    (most detailed) molecular functions in particular 
    from two sets of GO terms.

    :arg goa1: the first set of GO terms
    :type goa1: tuple, list, :class:`~numpy.ndarray`

    :arg goa2: the second set of GO terms
    :type goa2: tuple, list, :class:`~numpy.ndarray`
    """
    return_funcs = kwargs.pop('return_funcs', False)

    deepFuncs = [findDeepestFunctions(entry, **kwargs) for entry in goa_data]
    for i, entry in enumerate(deepFuncs):
        if len(entry) == 0:
            LOGGER.warn(
                'ensemble member {0} has no deep molecular functions and was omitted'
                .format(goa_data[i]._title))

    deepFuncs = [entry for entry in deepFuncs if len(entry) > 0]
    overlaps = calcGoOverlap(*deepFuncs, **kwargs)

    if return_funcs:
        return overlaps, deepFuncs

    return overlaps


def findDeepestFunctions(go_terms, **kwargs):
    """Find the deepest (most detailed) molecular functions in 
    a list of GO terms.

    :arg go_terms: a list of GO terms
    :type go_terms: :class:`.GOADictList`
    """
    if not isinstance(go_terms, GOADictList):
        raise TypeError('go_terms should be a GOADictList')

    go = kwargs.pop('go', None)
    if go is None:
        go = parseOBO(**kwargs)

    deep_functions = []
    for function in go_terms.getMolecularFunctions():
        if len(function.children) == 0:
            deep_functions.append(function)

    return deep_functions


def calcEnsembleFunctionOverlaps(ens, **kwargs):
    """Calculate function overlaps for an ensemble as the 
    mean of the value from :func:`calcDeepFunctionOverlaps`.

    :arg ens: an ensemble with labels
    :type ens: :class:`Ensemble`
    """
    if not isinstance(ens, Ensemble) and not isListLike(ens):
        raise TypeError('ens should be an ensemble or list-like')

    if isinstance(ens, Ensemble):
        ids = [label[:5] for label in ens.getLabels()]
    else:
        ids = ens

    if not isinstance(ids[0], str):
        raise TypeError('ens should have labels')

    goa_ens = queryGOA(ids, **kwargs)
    for entry in goa_ens:
        if len(entry._molecular) == 0:
            LOGGER.warn(
                'ensemble member {0} has no molecular functions and was omitted'.format(entry._title))

    goa_ens = [entry for entry in goa_ens if len(entry._molecular) > 0]

    overlaps = calcDeepFunctionOverlaps(*goa_ens, **kwargs)

    return overlaps



# -*- coding: utf-8 -*-
"""ProDy test suite.  Usage::

  from prody import *
  prody.test()

Testing will use :mod:`nose` if it is available, otherwise it will use
:mod:`unittest`."""

from glob import glob
from os.path import abspath, split, join, relpath, splitext
from os.path import sep as dirsep
import inspect
import tempfile

try:
    import unittest2 as unittest
    from unittest2 import TestCase, skipIf, skipUnless
except ImportError:
    import unittest
    from unittest import TestCase, skipIf, skipUnless

from prody.utilities import PLATFORM
from prody import LOGGER

from prody.utilities import which
NOPRODYCMD = which('prody') is None

WINDOWS = PLATFORM == 'Windows'

try:
    import matplotlib
    matplotlib.use('Agg')
except ImportError:
    MATPLOTLIB = False
else:
    try:
        from matplotlib import pyplot
    except ImportError:
        MATPLOTLIB = False
    else:
        MATPLOTLIB = True

TESTDIR = abspath(split(inspect.getfile(inspect.currentframe()))[0])
TEMPDIR = tempfile.gettempdir()


# -*- coding: utf-8 -*-
"""This module defines a class handling normal mode analysis data."""

import numpy as np

from .mode import Mode
from .modeset import ModeSet

from prody import PY2K, LOGGER
from prody.utilities import isListLike

if PY2K:
    range = xrange

__all__ = ['NMA', 'MaskedNMA']

class NMA(object):

    """A class for handling Normal Mode Analysis (NMA) data."""

    def __init__(self, title='Unknown'):

        self._title = str(title).strip()
        self._n_modes = 0
        self._cov = None
        self._n_atoms = 0
        self._dof = 0
        self._array = None      # modes/eigenvectors
        self._eigvals = None
        self._vars = None       # evs for PCA, inverse evs for ENM
        self._trace = None
        self._is3d = True       # is set to false for GNM
        self._indices = None

    def __len__(self):

        return self._n_modes

    def __getitem__(self, index):
        """A list or tuple of integers can be used for indexing."""

        if self._n_modes == 0:
            raise ValueError('{0} modes are not calculated, use '
                             'calcModes() method'.format(str(self)))
        if isinstance(index, slice):
            if index.stop is not None and index.stop > self.numModes():
                LOGGER.warn('The selection index contains {0}, a higher number than the total mode number ({1})'.format(index.stop, self.numModes()))
            if  index.start is not None and index.start > self.numModes()-1:
                raise IndexError('The selection index starts at {0}, a higher number than the highest mode index ({1})'.format(index.start, self.numModes()-1))
            indices = np.arange(*index.indices(len(self)))
            if len(indices) > 0:
                return ModeSet(self, indices)
        elif isinstance(index, (list, tuple, np.ndarray)):
            if len(index) == 1:
                return self._getMode(index[0])
            return ModeSet(self, index)
        try:
            index = int(index)
        except Exception:
            raise IndexError('indices must be int, slice, list, tuple or array')
        else:
            return self._getMode(index)

    def __iter__(self):

        for i in range(self._n_modes):
            yield self[i]

    def __repr__(self):

        return '<{0}: {1} ({2} modes; {3} atoms)>'.format(
                self.__class__.__name__, self._title, self._n_modes,
                self._n_atoms)

    def __str__(self):

        return self.__class__.__name__ + ' ' + self._title

    def _reset(self):
        self._n_modes = 0
        self._cov = None

        self._n_atoms = 0
        self._dof = 0

        self._array = None
        self._eigvals = None
        self._vars = None
        self._trace = None

        self._is3d = True

    def _clear(self):
        pass

    def _getMode(self, index):

        if self._n_modes == 0:
            raise ValueError('{0} modes are not calculated, use '
                             'calcModes() method'.format(str(self)))
        if index >= self._n_modes or index < -self._n_modes:
            raise IndexError('{0} contains {1} modes, try 0 <= index < '
                             '{1}'.format(str(self), self._n_modes))
        if index < 0:
            index += self._n_modes
        return Mode(self, index)

    def _getTrace(self):
        """Returns trace, and emit a warning message if trace is calculated
        using eigenvalues of a subset of variances (eigenvalues or inverse
        eigenvalues)."""

        trace = self._trace
        if trace is None:
            if self._vars is None:
                raise ValueError('variances are not set or calculated')
            trace = self._vars.sum()
            diff = self._dof - self._n_modes
            if self._is3d and diff > 6:
                diff = True
            elif diff > 1:
                diff = True
            else:
                diff = False
            if diff:
                from prody import LOGGER
                LOGGER.warn('Total variance for {0} is calculated using '
                            '{1} available modes out of {2} possible.'
                            .format(str(self), self._n_modes, self._dof))
        return trace

    def getModel(self):
        """Returns self."""

        return self

    def is3d(self):
        """Returns **True** if model is 3-dimensional."""

        return self._is3d

    def numAtoms(self):
        """Returns number of atoms."""

        return self._n_atoms

    def numModes(self):
        """Returns number of modes in the instance (not necessarily maximum
        number of possible modes)."""

        return self._n_modes

    def numDOF(self):
        """Returns number of degrees of freedom."""

        return self._dof

    def numEntries(self):
        """Returns number of entries in one eigenvector."""
        
        arr = self._getArray()
        if arr is None:
            return 0
        return arr.shape[0]

    def getTitle(self):
        """Returns title of the model."""

        return self._title

    def setTitle(self, title):
        """Set title of the model."""

        self._title = str(title)

    def getEigvals(self):
        """Returns eigenvalues.  For :class:`.PCA` and :class:`.EDA` models
        built using coordinate data in Å, unit of eigenvalues is |A2|.  For
        :class:`.ANM`, :class:`.GNM`, and :class:`.RTB`, on the other hand,
        eigenvalues are in arbitrary or relative units but they correlate with
        stiffness of the motion along associated eigenvector."""

        if self._eigvals is None: return None
        return self._eigvals.copy()

    def getVariances(self):
        """Returns variances.  For :class:`.PCA` and :class:`.EDA` models
        built using coordinate data in Å, unit of variance is |A2|.  For
        :class:`.ANM`, :class:`.GNM`, and :class:`.RTB`, on the other hand,
        variance is the inverse of the eigenvalue, so it has arbitrary or
        relative units."""

        if self._vars is None: return None
        return self._vars.copy()

    def getArray(self):
        """Returns a copy of eigenvectors array."""

        if self._array is None: return None
        return self._getArray().copy()

    getEigvecs = getArray

    def _getArray(self):
        """Returns eigenvectors array."""

        if self._array is None: return None
        return self._array

    def getCovariance(self):
        """Returns covariance matrix.  If covariance matrix is not set or yet
        calculated, it will be calculated using available modes."""

        if self._cov is None:
            array = self.getArray()
            if array is None:
                return None
            self._cov = np.dot(array, np.dot(np.diag(self._vars), array.T))
        return self._cov

    def calcModes(self):
        """"""

    def addEigenpair(self, vector, value=None):
        """Add eigen *vector* and eigen *value* pair(s) to the instance.
        If eigen *value* is omitted, it will be set to 1.  Inverse
        eigenvalues are set as variances."""

        try:
            ndim, shape = vector.ndim, vector.shape
        except AttributeError:
            raise TypeError('vector must be a numpy array')

        if ndim > 2:
            raise ValueError('vector must be 1- or 2-dimensional')
        elif ndim == 2 and shape[0] < shape[1]:
            raise ValueError('eigenvectors must correspond to vector columns')
        else:
            vector = vector.reshape((shape[0], 1))

        eigval = value

        if eigval is None:
            if ndim == 1:
                eigval = np.ones(1)
            else:
                eigval = np.ones(shape[2])
        else:
            try:
                ndim2, shape2 = eigval.ndim, eigval.shape
            except AttributeError:
                try:
                    eigval = np.array([value], float)
                except Exception:
                    raise TypeError('value must be a number or array')
            else:
                if value.ndim > 1:
                    raise ValueError('value must be a 1-dimensional array')
                elif value.shape[0] != value.shape[0]:
                    raise ValueError('number of eigenvectors and eigenvalues '
                                     'must match')

        if self._array is None:
            return self.setEigens(vector, value)

        if vector.shape[0] != self._array.shape[0]:
            raise ValueError('shape of vector do not match shape of '
                             'existing eigenvectors')
        self._array = np.concatenate((self._array, vector), 1)
        self._eigvals = np.concatenate((self._eigvals, value))
        self._n_modes += shape[1]
        self._vars = 1 / self._eigvals

    def setEigens(self, vectors, values=None):
        """Set eigen *vectors* and eigen *values*.  If eigen *values* are
        omitted, they will be set to 1.  Inverse eigenvalues are set as
        variances."""

        try:
            ndim, shape = vectors.ndim, vectors.shape
        except AttributeError:
            raise TypeError('vectors must be a numpy array')

        if ndim != 2:
            raise ValueError('vectors must be a 2-dimensional array')
        else:
            dof = shape[0]
            if self._is3d:
                n_atoms = dof // 3
            else:
                n_atoms = dof
            if self._n_atoms > 0 and n_atoms != self._n_atoms:
                LOGGER.warn('the number of atoms of the model will be changed to {0}'
                            .format(n_atoms))
            n_modes = vectors.shape[1]
        if values is not None:
            if not isinstance(vectors, np.ndarray):
                raise TypeError('values must be a numpy.ndarray, not {0}'
                                .format(type(vectors)))
            elif values.ndim != 1:
                raise ValueError('values must be a 1-dimensional array')
            else:
                if values.shape[0] != vectors.shape[1]:
                    raise ValueError('number of vectors and values do not match')
        else:
            values = np.ones(n_modes)

        self._array = vectors
        self._eigvals = values
        self._dof = dof
        self._n_atoms = n_atoms
        self._n_modes = n_modes
        self._vars = 1 / values

        self._clear()

    def getIndices(self):
      if self._indices is not None:
        return self._indices
      else:
        return np.arange(self.numModes())

class MaskedNMA(NMA):
    def __init__(self, name='Unknown', mask=False, masked=True):
        if isinstance(name, str):
            super(MaskedNMA, self).__init__(name)
        else:
            model = name
            name = model.getTitle()
            super(MaskedNMA, self).__init__(name)
            self.__dict__.update(model.__dict__)
            
        self.mask = False
        self.masked = masked
        self._maskedarray = None

        if not np.isscalar(mask):
            self.mask = np.array(mask)

    def _isOriginal(self):
        return self.masked or np.isscalar(self.mask)

    def __repr__(self):
        if self._isOriginal():
            n_dummies = 0
        else:
            n_dummies = len(self.mask) - self._n_atoms

        return ('<{0}: {1} ({2} modes; {3} nodes + {4} dummies)>'
                .format(self.__class__.__name__, self._title, self.__len__(), 
                        self._n_atoms, n_dummies))

    def numAtoms(self):
        """Returns number of atoms."""

        if self._isOriginal():
            return self._n_atoms
        else:
            return len(self.mask)

    def _extend(self, arr, axis=None, defval=0):
        mask = self.mask#.copy()
        if self.is3d():
            mask = np.repeat(mask, 3)

        n_true = np.sum(mask)
        N = len(mask)

        if axis is None:
            axes = [i for i in range(arr.ndim)]
        elif not isListLike(axis):
            axes = [axis]
        else:
            axes = axis

        shape = np.array(arr.shape)
        shape[axes] = N

        whole_array = np.empty(shape, dtype=arr.dtype)
        whole_array.fill(defval)

        I = [np.arange(s) for s in shape]
        J = [np.arange(s) for s in arr.shape]

        for ax in axes:
            I[ax] = mask
            J[ax] = np.arange(n_true)

        whole_array[np.ix_(*I)] = arr[np.ix_(*J)]

        return whole_array

    def _getArray(self):
        """Returns eigenvectors array. The function returns 
        a copy of the array if *masked* is **True**."""

        if self._array is None: return None

        if self._isOriginal():
            array = self._array
        else:
            if self._maskedarray is None:
                array = self._maskedarray = self._extend(self._array, axis=0)
            else:
                array = self._maskedarray

        return array

    def setNumAtoms(self, n_atoms):
        """Fixes the tail of the model. If *n_atoms* is greater than the original size 
        (number of nodes), then extra hidden nodes will be added to the model, and if 
        not, the model will be trimmed so that the total number of nodes, including the 
        hidden ones, will be equal to the *n_atoms*. Note that if *masked* is **True**, 
        the *n_atoms* should be the number of *visible* nodes.

        :arg n_atoms: number of atoms of the model 
        :type n_atoms: int
        """
        def _fixLength(vector, length, filled_value=0, axis=0):
            shape = vector.shape
            dim = len(shape)

            if shape[axis] < length:
                dl = length - shape[0]
                pad_width = []
                for i in range(dim):
                    if i == axis:
                        pad_width.append((0, dl))
                    else:
                        pad_width.append((0, 0))
                vector = np.pad(vector, pad_width, 'constant', constant_values=(filled_value,))
            elif shape[axis] > length:
                vector = vector[:length]
            return vector

        trimmed = self.masked
        if trimmed:
            trimmed_length = self.numAtoms()
            self.masked = False
            extra_length = self.numAtoms() - trimmed_length
            n_atoms += extra_length

        if np.isscalar(self.mask):
            self.mask = np.ones(self.numAtoms(), dtype=bool)

        self.mask = _fixLength(self.mask, n_atoms, False)
        self.masked = trimmed
        self._n_atoms = np.sum(self.mask, dtype=int)
        return

    def setEigens(self, vectors, values=None):
        mask = self.mask
        if not self.masked:
            if not np.isscalar(mask):
                if self.is3d():
                    mask = np.repeat(mask, 3)
                try:
                    vectors = vectors[mask, :]
                except IndexError:
                    raise IndexError('size mismatch between vectors (%d) and mask (%d).'
                                     'Try set masked to False or reset mask'%(vectors.shape[0], len(mask)))
        self._maskedarray = None
        super(MaskedNMA, self).setEigens(vectors, values)

    def calcModes(self):
        self._maskedarray = None
        super(MaskedNMA, self).calcModes()

    def _reset(self):
        super(MaskedNMA, self)._reset()
        self._maskedarray = None



"""Perform GNM calculations and output the results in plain text NMD, and
graphical formats."""

from ..apptools import *
from .nmaoptions import *
from . import nmaoptions

__all__ = ['prody_gnm']

DEFAULTS = {}
HELPTEXT = {}
for key, txt, val in [
    ('model', 'index of model that will be used in the calculations', 1),
    ('altloc', 'alternative location identifiers for residues used in the calculations', "A"),
    ('cutoff', 'cutoff distance (A)', 10.),
    ('gamma', 'spring constant', '1.'),
    ('zeros', 'calculate zero modes', False),

    ('outbeta', 'write beta-factors calculated from GNM modes', False),
    ('kirchhoff', 'write Kirchhoff matrix', False),
    ('figcmap', 'save contact map (Kirchhoff matrix) figure', False),
    ('figbeta', 'save beta-factors figure', False),
    ('figmode', 'save mode shape figures for specified modes, '
                 'e.g. "1-3 5" for modes 1, 2, 3 and 5', '')]:

    DEFAULTS[key] = val
    HELPTEXT[key] = txt

DEFAULTS.update(nmaoptions.DEFAULTS)
HELPTEXT.update(nmaoptions.HELPTEXT)

DEFAULTS['prefix'] = '_gnm'

def prody_gnm(pdb, **kwargs):
    """Perform GNM calculations for *pdb*.

    """

    for key in DEFAULTS:
        if not key in kwargs:
            kwargs[key] = DEFAULTS[key]

    from os.path import isdir, join, exists
    outdir = kwargs.get('outdir')
    if not isdir(outdir):
        raise IOError('{0} is not a valid path'.format(repr(outdir)))

    import numpy as np
    import prody
    LOGGER = prody.LOGGER

    selstr = kwargs.get('select')
    prefix = kwargs.get('prefix')
    cutoff = kwargs.get('cutoff')
    nmodes = kwargs.get('nmodes')
    selstr = kwargs.get('select')
    model = kwargs.get('model')
    altloc = kwargs.get('altloc')
    zeros = kwargs.get('zeros')
    membrane = kwargs.get('membrane')

    if membrane and not exists(pdb):
        pdb = prody.fetchPDBfromOPM(pdb)

    pdb = prody.parsePDB(pdb, model=model, altloc=altloc)
    if prefix == '_gnm':
        prefix = pdb.getTitle() + '_gnm'

    select = pdb.select(selstr)
    if select is None:
        raise ValueError('selection {0} do not match any atoms'
                          .format(repr(selstr)))
    LOGGER.info('{0} atoms will be used for GNM calculations.'
                .format(len(select)))

    if membrane:
        gnm = prody.exGNM(pdb.getTitle())
    else:
        gnm = prody.GNM(pdb.getTitle())
    try:
        gamma = float(kwargs.get('gamma'))
        LOGGER.info("Using gamma {0}".format(gamma))
    except ValueError:
        try:
            Gamma = eval('prody.' + kwargs.get('gamma'))
            gamma = Gamma(select)
            LOGGER.info("Using gamma {0}".format(Gamma))
        except NameError:
            raise NameError("Please provide gamma as a float or ProDy Gamma class")
        except TypeError:
            raise TypeError("Please provide gamma as a float or ProDy Gamma class")

    nproc = kwargs.get('nproc')
    gnm.buildKirchhoff(select, cutoff, gamma)
    gnm.calcModes(nmodes, zeros=zeros, nproc=nproc)

    LOGGER.info('Writing numerical output.')

    if kwargs.get('outnpz'):
        prody.saveModel(gnm, join(outdir, prefix), 
                        matrices=kwargs.get('npzmatrices'))

    if kwargs.get('outscipion'):
        prody.writeScipionModes(outdir, gnm)

    prody.writeNMD(join(outdir, prefix + '.nmd'), gnm, select)

    extend = kwargs.get('extend')
    if extend:
        if extend == 'all':
            extended = prody.extendModel(gnm, select, pdb)
        else:
            extended = prody.extendModel(gnm, select, select | pdb.bb)
        prody.writeNMD(join(outdir, prefix + '_extended_' +
                       extend + '.nmd'), *extended)

    outall = kwargs.get('outall')
    delim = kwargs.get('numdelim')
    ext = kwargs.get('numext')
    format = kwargs.get('numformat')

    if outall or kwargs.get('outeig'):
        prody.writeArray(join(outdir, prefix + '_evectors'+ext),
                         gnm.getArray(), delimiter=delim, format=format)
        prody.writeArray(join(outdir, prefix + '_evalues'+ext),
                         gnm.getEigvals(), delimiter=delim, format=format)

    if outall or kwargs.get('outbeta'):
        from prody.utilities import openFile
        fout = openFile(prefix + '_beta'+ext, 'w', folder=outdir)
        fout.write('{0[0]:1s} {0[1]:4s} {0[2]:4s} {0[3]:5s} {0[4]:5s}\n'
                       .format(['C', 'RES', '####', 'Exp.', 'The.']))
        for data in zip(select.getChids(), select.getResnames(),
                        select.getResnums(), select.getBetas(),
                        prody.calcTempFactors(gnm, select)):
            fout.write('{0[0]:1s} {0[1]:4s} {0[2]:4d} {0[3]:5.2f} {0[4]:5.2f}\n'
                       .format(data))
        fout.close()

    if outall or kwargs.get('outcov'):
        prody.writeArray(join(outdir, prefix + '_covariance'+ext),
                         gnm.getCovariance(), delimiter=delim, format=format)

    if outall or kwargs.get('outcc') or kwargs.get('outhm'):
        cc = prody.calcCrossCorr(gnm)
        if outall or kwargs.get('outcc'):
            prody.writeArray(join(outdir, prefix + '_cross-correlations' +
                             ext), cc, delimiter=delim, format=format)
        if outall or kwargs.get('outhm'):
            prody.writeHeatmap(join(outdir, prefix + '_cross-correlations.hm'),
                               cc, resnum=select.getResnums(),
                               xlabel='Residue', ylabel='Residue',
                               title=gnm.getTitle() + ' cross-correlations')

    if outall or kwargs.get('kirchhoff'):
        prody.writeArray(join(outdir, prefix + '_kirchhoff'+ext),
                         gnm.getKirchhoff(), delimiter=delim, format=format)

    if outall or kwargs.get('outsf'):
        prody.writeArray(join(outdir, prefix + '_sqfluct'+ext),
                         prody.calcSqFlucts(gnm), delimiter=delim,
                         format=format)

    figall = kwargs.get('figall')
    cc = kwargs.get('figcc')
    sf = kwargs.get('figsf')
    bf = kwargs.get('figbeta')
    cm = kwargs.get('figcmap')
    modes = kwargs.get('figmode')

    if figall or cc or sf or bf or cm or modes:
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            LOGGER.warning('Matplotlib could not be imported. '
                           'Figures are not saved.')
        else:
            prody.SETTINGS['auto_show'] = False
            LOGGER.info('Saving graphical output.')
            format = kwargs.get('figformat')
            width = kwargs.get('figwidth')
            height = kwargs.get('figheight')
            dpi = kwargs.get('figdpi')
            format = format.lower()

            if figall or cc:
                plt.figure(figsize=(width, height))
                prody.showCrossCorr(gnm, interactive=False)
                plt.savefig(join(outdir, prefix + '_cc.'+format),
                    dpi=dpi, format=format)
                plt.close('all')

            if figall or cm:
                plt.figure(figsize=(width, height))
                prody.showContactMap(gnm, interactive=False)
                plt.savefig(join(outdir, prefix + '_cm.'+format),
                    dpi=dpi, format=format)
                plt.close('all')

            if figall or sf:
                plt.figure(figsize=(width, height))
                prody.showSqFlucts(gnm)
                plt.savefig(join(outdir, prefix + '_sf.'+format),
                    dpi=dpi, format=format)
                plt.close('all')

            if figall or bf:
                plt.figure(figsize=(width, height))
                bexp = select.getBetas()
                bcal = prody.calcTempFactors(gnm, select)
                plt.plot(bexp, label='Experimental')
                plt.plot(bcal, label=('Theoretical (corr coef = {0:.2f})'
                                        .format(np.corrcoef(bcal, bexp)[0,1])))
                plt.legend(prop={'size': 10})
                plt.xlabel('Node index')
                plt.ylabel('Experimental B-factors')
                plt.title(pdb.getTitle() + ' B-factors')
                plt.savefig(join(outdir, prefix + '_bf.'+format),
                    dpi=dpi, format=format)
                plt.close('all')

            if modes:
                indices = []
                items = modes.split()
                items = sum([item.split(',') for item in items], [])
                for item in items:
                    try:
                        item = item.split('-')
                        if len(item) == 1:
                            indices.append(int(item[0])-1)
                        elif len(item) == 2:
                            indices.extend(list(range(int(item[0])-1, int(item[1]))))
                    except:
                        pass
                for index in indices:
                    try:
                        mode = gnm[index]
                    except:
                        pass
                    else:
                        plt.figure(figsize=(width, height))
                        prody.showMode(mode)
                        plt.grid()
                        plt.savefig(join(outdir, prefix + '_mode_' +
                            str(mode.getIndex()+1) + '.' + format),
                            dpi=dpi, format=format)
                        plt.close('all')


_ = list(HELPTEXT)
_.sort()
for key in _:

    prody_gnm.__doc__ += """
    :arg {0}: {1}, default is ``{2!r}``""".format(key, HELPTEXT[key],
                                                  DEFAULTS[key])


def addCommand(commands):

    subparser = commands.add_parser('gnm',
        help='perform Gaussian network model calculations')

    subparser.add_argument('--quiet', help="suppress info messages to stderr",
        action=Quiet, nargs=0)

    subparser.add_argument('--examples', action=UsageExample, nargs=0,
        help='show usage examples and exit')

    subparser.set_defaults(usage_example=
    """This command performs GNM calculations for given PDB structure and \
outputs results in NMD format. If an identifier is passed, structure file \
will be downloaded from the PDB FTP server.

Fetch PDB 1p38, run GNM calculations using default parameters, and results:

  $ prody gnm 1p38

Fetch PDB 1aar, run GNM calculations with cutoff distance 7 angstrom for \
chain A carbon alpha atoms with residue numbers less than 70, and \
save all of the graphical output files:

  $ prody gnm 1aar -c 7 -s "calpha and chain A and resnum < 70" -A""",
    test_examples=[0, 1])

    group = addNMAParameters(subparser)

    group.add_argument('-c', '--cutoff', dest='cutoff', type=float,
        default=DEFAULTS['cutoff'], metavar='FLOAT',
        help=HELPTEXT['cutoff'] + ' (default: %(default)s)')

    group.add_argument('-g', '--gamma', dest='gamma', type=str,
        default=DEFAULTS['gamma'], metavar='STR',
        help=HELPTEXT['gamma'] + ' (default: %(default)s)')

    group.add_argument('-m', '--model', dest='model', type=int,
        metavar='INT', default=DEFAULTS['model'], help=HELPTEXT['model'])

    group.add_argument('-L', '--altloc', dest='altloc', type=str,
        metavar='INT', default=DEFAULTS['altloc'], help=HELPTEXT['altloc'])

    group.add_argument('-w', '--zero-modes', dest='zeros', action='store_true',
        default=DEFAULTS['zeros'], help=HELPTEXT['zeros'])

    group = addNMAOutput(subparser)

    group.add_argument('-b', '--beta-factors', dest='outbeta',
        action='store_true', default=DEFAULTS['outbeta'],
        help=HELPTEXT['outbeta'])

    group.add_argument('-k', '--kirchhoff', dest='kirchhoff',
        action='store_true',
        default=DEFAULTS['kirchhoff'], help=HELPTEXT['kirchhoff'])

    group = addNMAOutputOptions(subparser, '_gnm')

    group = addNMAFigures(subparser)

    group.add_argument('-B', '--beta-factors-figure', dest='figbeta',
        action='store_true',
        default=DEFAULTS['figbeta'], help=HELPTEXT['figbeta'])

    group.add_argument('-K', '--contact-map', dest='figcmap',
        action='store_true',
        default=DEFAULTS['figcmap'], help=HELPTEXT['figcmap'])

    group.add_argument('-M', '--mode-shape-figure', dest='figmode', type=str,
        default=DEFAULTS['figmode'], help=HELPTEXT['figmode'], metavar='STR')

    group = addNMAFigureOptions(subparser)

    subparser.add_argument('pdb', help='PDB identifier or filename')

    subparser.set_defaults(func=lambda ns: prody_gnm(ns.__dict__.pop('pdb'),
                                                     **ns.__dict__))
    subparser.set_defaults(subparser=subparser)



# -*- coding: utf-8 -*-
"""This module defines :class:`Donor` for dealing with donor information provided
by using :meth:`.AtomGroup.setDonors` method."""

from numbers import Integral
import numpy as np

__all__ = ['Donor']

class Donor(object):

    """A pointer class for donored atoms.  Following built-in functions are
    customized for this class:

    * :func:`len` returns donor length, i.e. :meth:`getLength`
    * :func:`iter` yields :class:`~.Atom` instances"""

    __slots__ = ['_ag', '_acsi', '_indices']

    def __init__(self, ag, indices, acsi=None):

        self._ag = ag
        self._indices = np.array(indices)
        if acsi is None:
            self._acsi = ag.getACSIndex()
        else:
            self._acsi = acsi

    def __repr__(self):

        one, two = self._indices
        names = self._ag._getNames()
        return '<Donor: {0}({1})--{2}({3}) from {4}>'.format(
                            names[one], one, names[two], two, str(self._ag))

    def __str__(self):

        one, two = self._indices
        names = self._ag._getNames()
        return '{0}({1})--{2}({3})'.format(
                                            names[one], one, names[two], two)

    def __eq__(self, other):

        return (isinstance(other, Donor) and other.getAtomGroup() is self._ag
                and (np.all(other.getIndices() == self._indices) or
                 np.all(other.getIndices() == list(reversed(self._indices)))))

    def __ne__(self, other):

        return not self.__eq__(other)

    def __len__(self):

        return self.getLength()

    def __iter__(self):

        for index in self._indices:
            yield self._ag[index]

    def getAtomGroup(self):
        """Returns atom group."""

        return self._ag

    def getAtoms(self):
        """Returns donored atoms."""

        return (self._ag[self._indices[0]], self._ag[self._indices[1]])

    def getIndices(self):
        """Returns indices of donored atoms."""

        return self._indices.copy()

    def getLength(self):
        """Returns donor length."""

        vector = self.getVector()
        return np.multiply(vector, vector, vector).sum() ** 0.5

    def getVector(self):
        """Returns donor vector that originates from the first atom."""

        one, two = self._indices
        acsi = self.getACSIndex()
        return self._ag._coords[acsi, two] - self._ag._coords[acsi, one]

    def getACSIndex(self):
        """Returns index of the coordinate set."""

        acsi = self._acsi
        if acsi >= self._ag._n_csets:
            raise ValueError('{0} has fewer coordsets than assumed by {1}'
                             .format(str(self._ag), str(self)))
        return acsi

    def setACSIndex(self, index):
        """Set the coordinate set at *index* active."""

        if self._ag._coords is None:
            raise AttributeError('coordinates are not set')

        if not isinstance(index, Integral):
            raise TypeError('index must be an integer')

        n_csets = self._ag._n_csets
        if n_csets <= index or n_csets < abs(index):
            raise IndexError('coordinate set index is out of range')

        if index < 0:
            index += n_csets

        self._acsi = index


def evalDonors(donors, n_atoms):
    """Returns an array mapping atoms to their common donor neighbors and an array
    that stores number of donors made by each atom.
    
    This excludes entries corresponding to missing hydrogens (with 0 in them that became -1)."""

    if donors.min() < 0:
        # remove entries corresponding to missing hydrogens (with 0 in them that became -1)
        donors = donors[np.argwhere(donors == -1)[-1][0]+1:]

    numdonors = np.bincount(donors.reshape((donors.shape[0] * 2)))        

    domap = np.zeros((n_atoms, numdonors.max()), int)
    index = np.zeros(n_atoms, int)

    domap.fill(-1)
    
    for donor in donors:
        a, b = donor
        domap[a, index[a]] = b
        domap[b, index[b]] = a
        index[donor] += 1
    return domap, numdonors


def trimDonors(donors, indices):
    """Returns donors between atoms at given indices."""

    iset = set(indices)
    donors = [donor for donor in donors if donor[0] in iset and donor[1] in iset]
    if donors:
        newindices = np.zeros(indices.max()+1, int)
        newindices[indices] = np.arange(len(indices))
        return newindices[np.array(donors)]



# -*- coding: utf-8 -*-
"""This module defines classes for handling residues."""

from .subset import AtomSubset
from .atom import Atom

__all__ = ['Residue']

class Residue(AtomSubset):

    """Instances of this class point to atoms with same residue numbers (and
    insertion codes) and are generated by :class:`.HierView` class.
    Following built-in functions are customized for this class:

    * :func:`len` returns the number of atoms in the instance.
    * :func:`iter` yields :class:`.Atom` instances.

    Indexing :class:`Residue` instances by *atom name* (:func:`str`), e.g.
    ``"CA"`` returns an :class:`.Atom` instance."""

    __slots__ = ['_ag', '_indices', '_hv', '_acsi', '_selstr']

    def __init__(self, ag, indices, hv, acsi=None, **kwargs):

        AtomSubset.__init__(self, ag, indices, acsi, **kwargs)
        self._hv = hv


    def __repr__(self):

        n_csets = self._ag.numCoordsets()
        chain = self.getChid()
        if chain is None:
            chain = ''
        else:
            chain = ' from Chain {0}'.format(self.getChid())

        if n_csets == 1:
            return ('<Residue: {0} {1}{2}{3} from {4} ({5} atoms)>'
                    ).format(self.getResname(), self.getResnum(),
                     self.getIcode() or '', chain, self._ag.getTitle(),
                     len(self))
        elif n_csets > 1:
            return ('<Residue: {0} {1}{2}{3} from {4} '
                    '({5} atoms; active #{6} of {7} coordsets)>').format(
                    self.getResname(), self.getResnum(), self.getIcode() or '',
                    chain, self._ag.getTitle(), len(self), self.getACSIndex(),
                    n_csets)
        else:
            return ('<Residue: {0} {1}{2}{3} from {4} ({5} atoms; '
                    'no coordinates)>').format(self.getResname(),
                    self.getResnum(), self.getIcode() or '', chain,
                    self._ag.getTitle(), len(self))

    def __str__(self):

        return '{0} {1}{2}'.format(self.getResname(), self.getResnum(),
                                         self.getIcode() or '')

    def getAtom(self, name):
        """Returns atom with given *name*, **None** if not found.  Assumes that
        atom names in the residue are unique.  If more than one atoms with the
        given *name* exists, the one with the smaller index will be returned.
        """

        acsi = self.getACSIndex()
        if isinstance(name, str):
            nz = (self.getNames() == name).nonzero()[0]
            if len(nz) > 0:
                return Atom(self._ag, self._indices[nz[0]], acsi)

    __getitem__ = getAtom

    def getChain(self):
        """Returns the chain that the residue belongs to."""

        chid = self.getChid()
        segname = self.getSegname()
        if chid is not None:
            return self._hv.getChain(chid, segname)

    def getResnum(self):
        """Returns residue number."""

        return self._ag._getResnums()[self._indices[0]]

    def setResnum(self, number):
        """Set residue number."""

        self.setResnums(number)

    def getResname(self):
        """Returns residue name."""

        data = self._ag._getResnames()
        if data is not None:
            return data[self._indices[0]]

    def setResname(self, name):
        """Set residue name."""

        self.setResnames(name)

    def getIcode(self):
        """Returns residue insertion code."""

        data = self._ag._getIcodes()
        if data is not None:
            return data[self._indices[0]]

    def setIcode(self, icode):
        """Set residue insertion code."""

        self.setIcodes(icode)

    def getResindex(self):
        """Returns residue index."""

        return self._ag._getResindices()[self._indices[0]]

    def getChid(self):
        """Returns chain identifier."""

        chids = self._ag._getChids()
        if chids is not None:
            return chids[self._indices[0]]

    def getSegname(self):
        """Returns segment name."""

        segnames = self._ag._getSegnames()
        if segnames is not None:
            return segnames[self._indices[0]]

    def getSegment(self):
        """Returns segment of the residue."""

        segname = self.getSegname()
        if segname is not None:
            return self._hv.getSegment(segname)

    def getSelstr(self):
        """Returns selection string that will select this residue."""

        icode = self.getIcode() or ''
        chain = self.getChain()
        if chain is None:
            if self._selstr:
                return 'resnum {0}{1} and ({1})'.format(
                            self.getResnum(), icode, self._selstr)
            else:
                return 'resnum {0}{1}'.format(self.getResnum(), icode)
        else:
            return 'resnum {0}{1} and ({2})'.format(
                                self.getResnum(), icode, chain.getSelstr())

    def getPrev(self):
        """Returns preceding residue in the atom group."""

        nextIndex = self.getResindex()-1
        if nextIndex < 0:
            return None
        return self._hv._getResidue(nextIndex)

    def getNext(self):
        """Returns following residue in the atom group."""

        return self._hv._getResidue(self.getResindex()+1)



from prody.tests import TestCase

from numpy.random import random
from numpy.testing import assert_equal

from prody.utilities import checkCoords, checkTypes

COORDS = random((10, 3))*10
COORDSET = random((2, 10, 3))*10

class TestCheckCoords(TestCase):

    def testInvalidCoords(self):

        self.assertRaises(TypeError, checkCoords, [None])

    def testCoords(self):

        self.assertTrue(checkCoords(COORDS))

    def testCoordset(self):

        self.assertTrue(checkCoords(COORDSET, csets=True))


    def testCoordsetNatoms(self):

        self.assertRaises(ValueError, checkCoords, COORDSET, csets=True,
                          natoms=20)


class testCheckTypes(TestCase):

    def testCorrectMonotypeOneArg(self):

        self.assertTrue(checkTypes({'i': 1}, i=int))

    def testCorrectMonotypeTwoArgs(self):

        self.assertTrue(checkTypes({'i': 1, 'n': 10}, i=int, n=int))

    def testCorrectMultitypeOneArg(self):

        self.assertTrue(checkTypes({'i': 1.}, i=(int, float)))

    def testCorrectMonotypeTwoArgs(self):

        self.assertTrue(checkTypes({'i': 1, 'n': 10.}, i=(int, float),
                                                       n=(int, float)))

    def testWrongMonotypeOneArg(self):

        self.assertRaises(TypeError, checkTypes, {'i': 1.0}, i=int)

    def testWrongMonotypeTwoArgs(self):

        self.assertRaises(TypeError, checkTypes, {'i': 1, 'n': 10},
                                                  i=int, n=int)

    def testWrongMultitypeOneArg(self):

        self.assertRaises(TypeError, checkTypes, {'i': '1.'},
                                      i=(int, float))

    def testWrongMonotypeTwoArgs(self):

        self.assertRaises(TypeError, checkTypes, {'i': 1, 'n': '10.'},
                            i=(int, float), n=(int, float))



# -*- coding: utf-8 -*-
"""This module defines functions for interfacing BioExcel-CV19 simulation database."""

__author__ = 'James Krieger'

from os.path import join, isfile, basename, splitext
from numbers import Number
import numpy as np

from prody import LOGGER, PY3K
from prody.utilities import makePath

from prody.atomic.atomgroup import AtomGroup
from prody.proteins.pdbfile import parsePDB
from prody.trajectory.psffile import parsePSF, writePSF
from prody.trajectory.dcdfile import parseDCD

__all__ = ['fetchBioexcelPDB', 'parseBioexcelPDB', 'convertXtcToDcd',
           'fetchBioexcelTrajectory', 'parseBioexcelTrajectory',
           'fetchBioexcelTopology', 'parseBioexcelTopology']

cv19_prefix = 'https://bioexcel-cv19.bsc.es/api/rest/v1/projects/'
mddb_prefix = 'https://irb.mddbr.eu/api/rest/v1/projects/'
mddb_dev_prefix = 'https://irb-dev.mddbr.eu/api/rest/v1/projects/'
dot_json_str = '.json'

def fetchBioexcelPDB(acc, **kwargs):
    """Returns a path to the downloaded PDB file corresponding
    to the starting structure from a BioExcel-CV19 trajectory.
    
    
    :arg acc: BioExcel-CV19 project accession or ID
    :type acc: str

    :arg timeout: timeout for blocking connection attempt in seconds, 
        default is 60
    :type timeout: int

    :arg folder: path to folder for storing the results
        default is the current working directory
    :type folder: str

    :arg outname: out filename, default is input ``'acc.pdb'``
    :type outname: str

    :arg selection: atom selection for download
        options are ``'_C'`` for all carbon atoms, 
        ``'backbone'`` for backbone atoms,
        or ``'backbone and _C'`` for both
    type selection: str

    :arg db: database to use, options are ``'mddb'``, ``'mddb-dev'`` and ``'cv19'``
        default is ``'mddb-dev'``
    :type db: str

    See https://bioexcel-cv19.bsc.es/api/rest/docs for more info
    """
    acc, _, selection, filepath, timeout, _ = checkInputs(acc, **kwargs)
    if not filepath.endswith('.pdb'):
        filepath += '.pdb'

    db = kwargs.get('db', 'mddb-dev')
    if db == 'cv19':
        prefix = cv19_prefix
    elif db == 'mddb-dev':
        prefix = mddb_dev_prefix
    else:
        prefix = mddb_prefix

    url = prefix + acc + "/structure"
    if selection is not None:
        url += '?selection=' + selection.replace(" ","%20")
    
    filepath = requestFromUrl(url, timeout, filepath, source='pdb', **kwargs)

    return filepath


def fetchBioexcelTrajectory(acc, **kwargs):
    """Returns a path to the downloaded BioExcel-CV19 trajectory
    in xtc format or in dcd format if mdtraj is installed and 
    *convert* is **True**.

    :arg acc: BioExcel-CV19 project accession or ID
    :type acc: str

    :arg timeout: timeout for blocking connection attempt in seconds, 
        default is 60
    :type timeout: int

    :arg folder: path to folder for storing the results
        default is the current working directory
    :type folder: str

    :arg outname: out filename, default is input ``'acc.pdb'``
    :type outname: str

    :arg frames: which frames to select in 
        e.g. ``'1-5,11-15'`` or ``'10:20:2'``
        default is to not specify and get all
    :type frames: str

    :arg selection: atom selection for download
        options are ``'_C'`` for all carbon atoms, 
        ``'backbone'`` for backbone atoms,
        or ``'backbone and _C'`` for both
    type selection: str

    :arg db: database to use, options are ``'mddb'``, ``'mddb-dev'`` and ``'cv19'``
        default is ``'mddb-dev'``
    :type db: str

    See https://bioexcel-cv19.bsc.es/api/rest/docs for more info

    :arg convert: convert to dcd if mdtraj is installed
        default is True
    type convert: bool
    """
    acc, convert, selection, filepath, timeout, frames = checkInputs(acc, **kwargs)
    if not filepath.endswith('.xtc'):
        filepath += '.xtc'

    db = kwargs.get('db', 'mddb-dev')
    if db == 'cv19':
        prefix = cv19_prefix
    elif db == 'mddb-dev':
        prefix = mddb_dev_prefix
    else:
        prefix = mddb_prefix

    url = prefix + acc + "/trajectory?format=xtc"

    if frames is not None:
        url += '&frames=' + frames

    if selection is not None:
        url += '&selection=' + selection.replace(" ","%20")

    filepath = requestFromUrl(url, timeout, filepath, source='xtc', **kwargs)

    if convert:
        filepath = convertXtcToDcd(filepath, **kwargs)

    return filepath


def fetchBioexcelTopology(acc, **kwargs):
    """Returns a path to the downloaded BioExcel-CV19 topology
    in json format or in psf format if *convert* is **True**.
    
    :arg acc: BioExcel-CV19 project accession or ID
    :type acc: str

    :arg timeout: timeout for blocking connection attempt in seconds, 
        default is 60
    :type timeout: int

    :arg folder: path to folder for storing the results
        default is the current working directory
    :type folder: str

    :arg outname: out filename, default is input ``'acc.pdb'``
    :type outname: str

    :arg db: database to use, options are ``'mddb'``, ``'mddb-dev'`` and ``'cv19'``
        default is ``'mddb-dev'``
    :type db: str

    See https://bioexcel-cv19.bsc.es/api/rest/docs for more info
    """
    if isfile(acc):
        filepath = acc
    else:
        acc, convert, _, filepath, timeout, _ = checkInputs(acc, **kwargs)
        if not filepath.endswith(dot_json_str) and not filepath.endswith('.psf'):
            filepath += dot_json_str

    if filepath.endswith('.psf'):
        convert = False

    db = kwargs.get('db', 'mddb-dev')
    if db == 'cv19':
        prefix = cv19_prefix
    elif db == 'mddb-dev':
        prefix = mddb_dev_prefix
    else:
        prefix = mddb_prefix

    if not isfile(filepath):
        url = prefix + acc + "/topology"
        filepath = requestFromUrl(url, timeout, filepath, source='json', **kwargs)

    if convert:
        ag = parseBioexcelTopology(filepath, **kwargs)
        filepath = filepath.replace(dot_json_str, '.psf')
        writePSF(filepath, ag)

    return filepath


def parseBioexcelTopology(query, **kwargs):
    """Parse a BioExcel-CV19 topology json into an :class:`.AtomGroup`,
    fetching it if needed using **kwargs
    """
    query = checkQuery(query)

    kwargs['convert'] = False
    if not isfile(query):
        filename = fetchBioexcelTopology(query, **kwargs)
    else:
        filename = query

    if filename.endswith(dot_json_str):
        import json

        fp = open(filename, 'r')
        data = json.load(fp)
        fp.close()

        title = basename(splitext(filename)[0])
        ag = AtomGroup(title)

        ag.setNames(data['atom_names'])
        ag.setElements(data['atom_elements'])
        ag.setCharges(data['atom_charges'])
        ag.setResnums(data['atom_residue_indices'])

        # set false n_csets and acsi to allow nodes in ag
        ag._n_csets = 1
        ag._acsi = 0

        indices = np.ix_(*[np.array(data['atom_residue_indices'])])

        chids = np.array([data['chain_names'][chain_index]
                          for chain_index in data['residue_chain_indices']])
        ag.setChids(chids[indices])

        ag.setResnames(np.array(data['residue_names'])[indices])
        ag.setResnums(np.array(data['residue_numbers'])[indices])

        if data['residue_icodes'] is not None:
            ag.setIcodes(np.array(data['residue_icodes'])[indices])

        # restore acsi and n_csets to defaults
        ag._acsi = None
        ag._n_csets = 0
    else:
        ag = parsePSF(filename)

    selection = checkSelection(**kwargs)

    if selection == '_C':
        ag = ag.select('element C').copy()
    elif selection == 'backbone':
        ag = ag.select('backbone').copy()
    elif selection == 'backbone and _C':
        ag = ag.select('backbone and element C')

    return ag

def parseBioexcelTrajectory(query, **kwargs):
    """Parse a BioExcel-CV19 topology json into an :class:`.Ensemble`,
    fetching it if needed using **kwargs

    :arg top: topology filename
    :type top: str
    """
    kwargs['convert'] = True
    if isfile(query) and query.endswith('.dcd'):
        filename = query
    elif isfile(query + '.dcd'):
        filename = query + '.dcd'
    elif isfile(query) and query.endswith('.xtc'):
        filename = convertXtcToDcd(query, **kwargs)
    elif isfile(query + '.xtc'):
        filename = convertXtcToDcd(query + '.xtc', **kwargs)
    else:
        filename = fetchBioexcelTrajectory(query, **kwargs)

    return parseDCD(filename)

def parseBioexcelPDB(query, **kwargs):
    """Parse a BioExcel-CV19 topology json into an :class:`.Ensemble`,
    fetching it if needed using **kwargs
    """
    kwargs['convert'] = True
    if isfile(query):
        filename = query
    elif isfile(query + '.pdb'):
        filename = query + '.pdb'
    else:
        filename = fetchBioexcelPDB(query, **kwargs)

    ag = parsePDB(filename)
    if ag is None:
        filename = fetchBioexcelPDB(query, **kwargs)
        ag = parsePDB(filename)

    acc = basename(splitext(filename)[0])
    ag2 = parseBioexcelTopology(acc, **kwargs)

    ag.setElements(ag2.getElements())
    return ag

def convertXtcToDcd(filepath, **kwargs):
    """Convert xtc trajectories to dcd files using mdtraj.
    Returns path to output dcd file.

    :arg top: topology filename
    :type top: str    
    """
    topFile = kwargs.get('top', None)
    if topFile is not None:
        acc = topFile
    else:
        acc = basename(splitext(filepath)[0])

    if not isfile(acc):
        acc = fetchBioexcelTopology(acc, **kwargs)

    try:
        import mdtraj
    except ImportError:
        raise ImportError('Please install mdtraj to convert to dcd.')
    else:
        top = mdtraj.load_topology(acc)
        traj = mdtraj.load_xtc(filepath, top=top)
        filepath = filepath.replace('xtc', 'dcd')
        traj.save_dcd(filepath)

    return filepath

def requestFromUrl(url, timeout, filepath, source=None, **kwargs):
    """Helper function to make a request from a url and return the response"""
    import requests
    import json
    import mdtraj
    import tempfile

    db = kwargs.get('db', 'mddb-dev')
    if db == 'cv19':
        prefix = cv19_prefix
    elif db == 'mddb-dev':
        prefix = mddb_dev_prefix
    else:
        prefix = mddb_prefix

    acc = url.split(prefix)[1].split('/')[0]

    LOGGER.timeit('_bioexcel')
    response = None
    sleep = 2
    while LOGGER.timing('_bioexcel') < timeout:
        try:
            response = requests.get(url).content

            if source == 'json':
                json.loads(response)

                if PY3K:
                    response = response.decode()

                fo = open(filepath, 'w')
                fo.write(response)
                fo.close()

            elif source == 'xtc':
                fo = open(filepath, 'wb')
                fo.write(response)
                fo.close()
                
                top = mdtraj.load_psf(fetchBioexcelTopology(acc, timeout=timeout))
                mdtraj.load_xtc(filepath, top=top)

            elif source == 'pdb':
                if PY3K:
                    response = response.decode()

                fo = open(filepath, 'w')
                fo.write(response)
                fo.close()

                ag = parsePDB(filepath)
                numAtoms = ag.numAtoms()

        except Exception:
            pass
        else:
            break
        
        sleep = 100 if int(sleep * 1.5) >= 100 else int(sleep * 1.5)
        LOGGER.sleep(int(sleep), '. Trying to reconnect...')

    return filepath

def checkSelection(**kwargs):
    """Helper function to check selection"""
    selection = kwargs.get('selection', None)
    if selection is not None:
        if not isinstance(selection, str):
            raise TypeError('selection should be a string')

        if selection not in ['_C', 'backbone', 'backbone and _C']:
            raise ValueError("selection should be '_C', 'backbone' or 'backbone and _C'")

    return selection

def checkQuery(input):
    """Check query or acc argument, which should be a string as follows:
    
    :arg acc: BioExcel-CV19 project accession or ID
    :type acc: str
    """
    if not isinstance(input, str):
        raise TypeError('query should be string')
    return input

def checkConvert(**kwargs):
    convert = kwargs.get('convert', True)
    if not isinstance(convert, (bool, type(None))):
        raise TypeError('convert should be bool')
    return convert

def checkTimeout(**kwargs):
    timeout = kwargs.get('timeout', 200)
    if not isinstance(timeout, (Number, type(None))):
        raise TypeError('timeout should be number')
    return timeout


def checkFilePath(query, **kwargs):
    """Check folder and outname and path, making it if needed"""

    folder = kwargs.get('folder', '.')
    if not isinstance(folder, str):
        try:
            folder = str(folder)
        except Exception:
            raise TypeError('folder should be a string')
    
    outname = kwargs.get('outname', None)
    if not outname:
        outname = query
    elif not isinstance(outname, str):
        raise TypeError('outname should be a string')

    return join(makePath(folder), outname)


def checkFrames(**kwargs):
    """Check frames kwarg matches the following:

    :arg frames: which frames to select in 
        e.g. ``'1-5,11-15'`` or ``'10:20:2'``
        default is to not specify and get all
    :type frames: str
    """
    frames = kwargs.get('frames', None)
    if frames is None:
        return frames
    
    if not isinstance(frames, str):
        raise TypeError('frames should be a string')
 
    for framesRange in frames.split(','):
        if framesRange.find('-') != -1 and framesRange.find(':') != -1:
            raise ValueError('Frames should have a set of comma-separated ranges containing either a hyphen or colons, not both')
        
        if framesRange.count('-') > 0:
            if not np.all([item.isnumeric() for item in framesRange.split('-')]):
                raise ValueError('Each frames range should only have numbers and hyphens (or colons), not spaces or anything else')
            if framesRange.count('-') > 1:
                raise ValueError('Each frames range can only have one hyphen')
        
        if framesRange.count(':') not in [0,1,2]:
            raise ValueError('Each frames range can have no more than 2 colons')

        if framesRange.count(':') > 0:
            if not np.all([item.isnumeric() for item in framesRange.split(':')]):
                raise ValueError('Each frames range should only have numbers and colons (or hyphens), not spaces or anything else')

        if (framesRange.find('-') == -1 and framesRange.find(':') == -1 and 
            not framesRange.isnumeric()):
            raise ValueError('Each frames range should only have numbers (and colons or hyphens), not spaces or anything else')

    return frames

def checkInputs(query, **kwargs):
    query = checkQuery(query)
    timeout = checkTimeout(**kwargs)
    convert = checkConvert(**kwargs)
    selection = checkSelection(**kwargs)
    filepath = checkFilePath(query, **kwargs)
    frames = checkFrames(**kwargs)
    return query, convert, selection, filepath, timeout, frames
